# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from datetime import datetime, timedelta
from typing import ClassVar

import toml

from airbyte_ops_mcp.connector_ops.utils import (
    Connector,
    ConnectorLanguage,
)  # type: ignore
from airbyte_ops_mcp.connector_qa import consts
from airbyte_ops_mcp.connector_qa.models import (
    Check,
    CheckCategory,
    CheckResult,
)
from airbyte_ops_mcp.metadata_validator import (
    PRE_UPLOAD_VALIDATORS,
    ValidatorOptions,
    validate_and_load,
)


class MetadataCheck(Check):
    category = CheckCategory.METADATA


class ValidateMetadata(MetadataCheck):
    name = f"Connectors must have valid {consts.METADATA_FILE_NAME} file"
    description = f"Connectors must have a `{consts.METADATA_FILE_NAME}` file at the root of their directory. This file is used to build our connector registry. Its structure must follow our metadata schema. Field values are also validated. This is to ensure that all connectors have the required metadata fields and that the metadata is valid. More details in this [documentation]({consts.METADATA_DOCUMENTATION_URL})."

    def _run(self, connector: Connector) -> CheckResult:
        deserialized_metadata, error = validate_and_load(
            connector.metadata_file_path,
            PRE_UPLOAD_VALIDATORS,
            ValidatorOptions(docs_path=str(connector.documentation_file_path)),
        )
        if not deserialized_metadata:
            return self.fail(
                connector=connector, message=f"Metadata file is invalid: {error}"
            )

        return self.pass_(
            connector=connector,
            message="Metadata file valid.",
        )


class CheckConnectorLanguageTag(MetadataCheck):
    name = "Connector must have a language tag in metadata"
    description = f"Connectors must have a language tag in their metadata. It must be set in the `tags` field in {consts.METADATA_FILE_NAME}. The values can be `language:python` or `language:java`. This checks infers the correct language tag based on the presence of certain files in the connector directory."

    PYTHON_LANGUAGE_TAG = "language:python"
    JAVA_LANGUAGE_TAG = "language:java"
    MANIFEST_ONLY_LANGUAGE_TAG = "language:manifest-only"

    def get_expected_language_tag(self, connector: Connector) -> str:
        if (connector.code_directory / "manifest.yaml").exists():
            return self.MANIFEST_ONLY_LANGUAGE_TAG
        if (connector.code_directory / consts.SETUP_PY_FILE_NAME).exists() or (
            connector.code_directory / consts.PYPROJECT_FILE_NAME
        ).exists():
            return self.PYTHON_LANGUAGE_TAG
        elif (connector.code_directory / consts.GRADLE_FILE_NAME).exists() or (
            connector.code_directory / consts.GRADLE_KOTLIN_FILE_NAME
        ).exists():
            return self.JAVA_LANGUAGE_TAG
        else:
            raise ValueError(
                "Could not infer the language tag from the connector directory"
            )

    def _run(self, connector: Connector) -> CheckResult:
        try:
            expected_language_tag = self.get_expected_language_tag(connector)
        except ValueError:
            return self.fail(
                connector=connector,
                message="Could not infer the language tag from the connector directory",
            )

        current_language_tags = [
            t for t in connector.metadata.get("tags", []) if t.startswith("language:")
        ]
        if not current_language_tags:
            return self.fail(
                connector=connector,
                message="Language tag is missing in the metadata file",
            )
        if len(current_language_tags) > 1:
            return self.fail(
                connector=connector,
                message=f"Multiple language tags found in the metadata file: {current_language_tags}",
            )
        current_language_tag = current_language_tags[0]
        if current_language_tag != expected_language_tag:
            return self.fail(
                connector=connector,
                message=f"Expected language tag '{expected_language_tag}' in the {consts.METADATA_FILE_NAME} file, but found '{current_language_tag}'",
            )
        return self.pass_(
            connector=connector,
            message=f"Language tag {expected_language_tag} is present in the metadata file",
        )


class CheckConnectorCDKTag(MetadataCheck):
    name = "Python connectors must have a CDK tag in metadata"
    description = f"Python connectors must have a CDK tag in their metadata. It must be set in the `tags` field in {consts.METADATA_FILE_NAME}. The values can be `cdk:low-code`, `cdk:python`, or `cdk:file`."
    applies_to_connector_languages: ClassVar[list] = [
        ConnectorLanguage.PYTHON,
        ConnectorLanguage.LOW_CODE,
    ]

    class CDKTag:
        LOW_CODE = "cdk:low-code"
        PYTHON = "cdk:python"
        FILE = "cdk:python-file-based"

    def get_expected_cdk_tag(self, connector: Connector) -> str:
        manifest_file = (
            connector.code_directory
            / connector.technical_name.replace("-", "_")
            / consts.LOW_CODE_MANIFEST_FILE_NAME
        )
        pyproject_file = connector.code_directory / consts.PYPROJECT_FILE_NAME
        setup_py_file = connector.code_directory / consts.SETUP_PY_FILE_NAME
        if manifest_file.exists():
            return self.CDKTag.LOW_CODE
        if pyproject_file.exists():
            pyproject = toml.load(connector.code_directory / consts.PYPROJECT_FILE_NAME)
            cdk_deps = pyproject["tool"]["poetry"]["dependencies"].get(
                "airbyte-cdk", None
            )
            if (
                cdk_deps
                and isinstance(cdk_deps, dict)
                and "file-based" in cdk_deps.get("extras", [])
            ):
                return self.CDKTag.FILE
        if setup_py_file.exists() and (
            "airbyte-cdk[file-based]"
            in (connector.code_directory / consts.SETUP_PY_FILE_NAME).read_text()
        ):
            return self.CDKTag.FILE
        return self.CDKTag.PYTHON

    def _run(self, connector: Connector) -> CheckResult:
        current_cdk_tags = [
            t for t in connector.metadata.get("tags", []) if t.startswith("cdk:")
        ]
        expected_cdk_tag = self.get_expected_cdk_tag(connector)
        if not current_cdk_tags:
            return self.fail(
                connector=connector,
                message="CDK tag is missing in the metadata file",
            )
        if len(current_cdk_tags) > 1:
            return self.fail(
                connector=connector,
                message=f"Multiple CDK tags found in the metadata file: {current_cdk_tags}",
            )
        if current_cdk_tags[0] != expected_cdk_tag:
            return self.fail(
                connector=connector,
                message=f"Expected CDK tag '{self.get_expected_cdk_tag(connector)}' in the {consts.METADATA_FILE_NAME} file, but found '{current_cdk_tags[0]}'",
            )
        return self.pass_(
            connector=connector,
            message=f"CDK tag {self.get_expected_cdk_tag(connector)} is present in the metadata file",
        )


class ValidateBreakingChangesDeadlines(MetadataCheck):
    """
    Verify that _if_ the most recent connector version has a breaking change,
    it's deadline is at least a week in the future.
    """

    name = "Breaking change deadline should be a week in the future"
    description = "If the connector version has a breaking change, the deadline field must be set to at least a week in the future. See the Breaking Changes Policy for full requirements: https://docs.airbyte.com/platform/connector-development/connector-breaking-changes"
    runs_on_released_connectors = False
    minimum_days_until_deadline = 7

    def _run(self, connector: Connector) -> CheckResult:
        # fetch the current branch version of the connector first.
        # we'll try and see if there are any breaking changes associated
        # with it next.
        current_version = connector.version
        if current_version is None:
            return self.fail(
                connector=connector,
                message="Can't verify breaking changes deadline: connector version is not defined.",
            )

        breaking_changes = connector.metadata.get("releases", {}).get("breakingChanges")

        if not breaking_changes:
            return self.pass_(
                connector=connector,
                message="No breaking changes found on this connector.",
            )

        current_version_breaking_changes = breaking_changes.get(current_version)

        if not current_version_breaking_changes:
            return self.pass_(
                connector=connector,
                message="No breaking changes found for the current version.",
            )

        upgrade_deadline = current_version_breaking_changes.get("upgradeDeadline")

        if not upgrade_deadline:
            return self.fail(
                connector=connector,
                message=f"No upgrade deadline found for the breaking changes in {current_version}.",
            )

        upgrade_deadline_datetime = datetime.strptime(upgrade_deadline, "%Y-%m-%d")
        one_week_from_now = datetime.utcnow() + timedelta(
            days=self.minimum_days_until_deadline
        )

        if upgrade_deadline_datetime <= one_week_from_now:
            return self.fail(
                connector=connector,
                message=f"The upgrade deadline for the breaking changes in {current_version} is less than {self.minimum_days_until_deadline} days from today. Please extend the deadline",
            )

        return self.pass_(
            connector=connector,
            message="The upgrade deadline is set to at least a week in the future",
        )


class CheckConnectorMaxSecondsBetweenMessagesValue(MetadataCheck):
    name = "Certified source connector must have a value filled out for maxSecondsBetweenMessages in metadata"
    description = "Certified source connectors must have a value filled out for `maxSecondsBetweenMessages` in metadata. This value represents the maximum number of seconds we could expect between messages for API connectors. And it's used by platform to tune connectors heartbeat timeout. The value must be set in the 'data' field in connector's `metadata.yaml` file."
    applies_to_connector_types: ClassVar[list[str]] = ["source"]
    applies_to_connector_support_levels: ClassVar[list[str]] = ["certified"]

    def _run(self, connector: Connector) -> CheckResult:
        max_seconds_between_messages = connector.metadata.get(
            "maxSecondsBetweenMessages"
        )
        if not max_seconds_between_messages:
            return self.fail(
                connector=connector,
                message="Missing required for certified connectors field 'maxSecondsBetweenMessages'",
            )
        return self.pass_(
            connector=connector,
            message="Value for maxSecondsBetweenMessages is set",
        )


ENABLED_CHECKS = [
    ValidateMetadata(),
    CheckConnectorLanguageTag(),
    CheckConnectorCDKTag(),
    ValidateBreakingChangesDeadlines(),
    CheckConnectorMaxSecondsBetweenMessagesValue(),
]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import hashlib
from typing import List

import pandas as pd
from dagster import MetadataValue, Output

OutputDataFrame = Output[pd.DataFrame]
CURSOR_SEPARATOR = ":"


def output_dataframe(result_df: pd.DataFrame) -> Output[pd.DataFrame]:
    """
    Returns a Dagster Output object with a dataframe as the result and a markdown preview.
    """

    # Truncate to 10 rows to avoid dagster throwing a "too large" error
    MAX_PREVIEW_ROWS = 10
    is_truncated = len(result_df) > MAX_PREVIEW_ROWS
    preview_result_df = result_df.head(MAX_PREVIEW_ROWS)

    return Output(
        result_df,
        metadata={
            "count": len(result_df),
            "preview": MetadataValue.md(preview_result_df.to_markdown()),
            "is_truncated": is_truncated,
        },
    )


def string_array_to_hash(strings: List[str]) -> str:
    """Hash a list of strings into a cursor string.

    Args:
        unique_strings (List[str]): unique strings

    Returns:
        str: A cursor string
    """
    unique_strings = list(set(strings))
    unique_strings.sort()
    return hashlib.md5(str(unique_strings).encode("utf-8")).hexdigest()



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

import copy
from typing import List, Optional, TypedDict

from typing_extensions import NotRequired


class Relationship(TypedDict):
    name: str
    relations: dict[str, str]
    false_positives: NotRequired[dict[str, str]]


class Relationships(TypedDict):
    streams: List[Relationship]


class RelationshipsMerger:
    def merge(
        self,
        estimated_relationships: Relationships,
        confirmed_relationships: Relationships,
    ) -> Relationships:
        streams = []
        for estimated_stream in estimated_relationships["streams"]:
            confirmed_relationships_for_stream = self._get_stream(
                confirmed_relationships, estimated_stream["name"]
            )
            if confirmed_relationships_for_stream:
                streams.append(
                    self._merge_for_stream(
                        estimated_stream, confirmed_relationships_for_stream
                    )
                )  # type: ignore  # at this point, we know confirmed_relationships_for_stream is not None
            else:
                streams.append(estimated_stream)

        already_processed_streams = set(
            map(lambda relationship: relationship["name"], streams)
        )
        for confirmed_stream in confirmed_relationships["streams"]:
            if confirmed_stream["name"] not in already_processed_streams:
                streams.append(
                    {
                        "name": confirmed_stream["name"],
                        "relations": confirmed_stream["relations"],
                    }
                )
        return {"streams": streams}

    def _merge_for_stream(
        self, estimated: Relationship, confirmed: Relationship
    ) -> Relationship:
        relations = copy.deepcopy(confirmed.get("relations", {}))

        # get estimated but filter out false positives
        for field, target in estimated.get("relations", {}).items():
            false_positives = (
                confirmed["false_positives"] if "false_positives" in confirmed else {}
            )
            if field not in relations and (
                field not in false_positives
                or false_positives.get(field, None) != target
            ):  # type: ignore  # at this point, false_positives should not be None
                relations[field] = target

        return {
            "name": estimated["name"],
            "relations": relations,
        }

    def _get_stream(
        self, relationships: Relationships, stream_name: str
    ) -> Optional[Relationship]:
        for stream in relationships["streams"]:
            if stream.get("name", None) == stream_name:
                return stream

        return None



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Configuration options for live tests.

This module provides configuration classes and enums for controlling
live test behavior, including connection filtering, stream selection,
and test modes.

Based on airbyte-ci implementation:
https://github.com/airbytehq/airbyte/blob/master/airbyte-ci/connectors/live-tests/src/live_tests/commons/models.py
https://github.com/airbytehq/airbyte/blob/master/airbyte-ci/connectors/live-tests/src/live_tests/commons/connection_objects_retrieval.py
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path


class ConnectionSubset(Enum):
    """Signals which connection pool to consider for live tests.

    SANDBOXES: Only use Airbyte sandbox connections (safer, limited data)
    ALL: Use all available connections on Cloud (more coverage, real data)
    """

    SANDBOXES = "sandboxes"
    ALL = "all"

    @classmethod
    def from_string(
        cls,
        value: str,
    ) -> ConnectionSubset:
        """Parse connection subset from string."""
        value_lower = value.lower()
        if value_lower == "sandboxes":
            return cls.SANDBOXES
        if value_lower == "all":
            return cls.ALL
        raise ValueError(
            f"Unknown connection subset: {value}. Must be 'sandboxes' or 'all'."
        )


class TargetOrControl(Enum):
    """Identifies whether a connector is the target or control version."""

    TARGET = "target"
    CONTROL = "control"


class ActorType(Enum):
    """Type of connector actor."""

    SOURCE = "source"
    DESTINATION = "destination"


@dataclass
class LiveTestConfig:
    """Configuration for live test execution.

    This class consolidates all configuration options for running live tests,
    including connection filtering, stream selection, and test behavior.
    """

    # Connection filtering
    connection_id: str | None = None
    connection_subset: ConnectionSubset = ConnectionSubset.SANDBOXES
    max_connections: int | None = None
    auto_select_connections: bool = False

    # Stream filtering
    selected_streams: set[str] | None = None

    # Custom paths for local testing
    custom_config_path: Path | None = None
    custom_catalog_path: Path | None = None
    custom_state_path: Path | None = None

    # Test behavior
    test_description: str | None = None
    retrieval_reason: str | None = None

    def __post_init__(self) -> None:
        """Validate configuration after initialization."""
        if self.connection_id and self.auto_select_connections:
            raise ValueError(
                "Cannot set both connection_id and auto_select_connections"
            )


@dataclass
class StreamFilter:
    """Filter for selecting which streams to test.

    Provides utilities for filtering streams based on include/exclude patterns.
    """

    include_streams: set[str] | None = None
    exclude_streams: set[str] | None = None

    def filter_streams(
        self,
        available_streams: set[str],
    ) -> set[str]:
        """Filter available streams based on include/exclude rules.

        If include_streams is set, only those streams are included.
        If exclude_streams is set, those streams are removed from the result.
        """
        if self.include_streams:
            result = available_streams & self.include_streams
        else:
            result = available_streams.copy()

        if self.exclude_streams:
            result = result - self.exclude_streams

        return result

    def matches(
        self,
        stream_name: str,
    ) -> bool:
        """Check if a stream name matches the filter."""
        if self.include_streams and stream_name not in self.include_streams:
            return False
        return not (self.exclude_streams and stream_name in self.exclude_streams)


@dataclass
class ConnectionCandidate:
    """Represents a candidate connection for testing.

    Used when auto-selecting connections to test based on stream coverage
    and sync duration.
    """

    connection_id: str
    workspace_id: str | None = None
    streams_with_data: list[str] = field(default_factory=list)
    last_sync_duration_seconds: float | None = None

    @property
    def stream_count(self) -> int:
        return len(self.streams_with_data)


def select_best_connection_candidates(
    candidates: list[ConnectionCandidate],
    max_connections: int | None = None,
) -> list[tuple[ConnectionCandidate, list[str]]]:
    """Select the best subset of connection candidates for testing.

    This function reduces the list of candidates to minimize the number of
    connections while maximizing stream coverage. It prioritizes faster
    connections (shorter sync duration).

    Based on airbyte-ci implementation:
    https://github.com/airbytehq/airbyte/blob/master/airbyte-ci/connectors/live-tests/src/live_tests/commons/connection_objects_retrieval.py#L201-L220
    """
    # Sort by sync duration (faster first)
    sorted_candidates = sorted(
        candidates,
        key=lambda c: c.last_sync_duration_seconds or float("inf"),
    )

    tested_streams: set[str] = set()
    selected: list[tuple[ConnectionCandidate, list[str]]] = []

    for candidate in sorted_candidates:
        streams_to_test = []
        for stream in candidate.streams_with_data:
            if stream not in tested_streams:
                streams_to_test.append(stream)
                tested_streams.add(stream)

        if streams_to_test:
            selected.append((candidate, streams_to_test))

    # Sort by number of streams (most streams first)
    selected = sorted(selected, key=lambda x: len(x[1]), reverse=True)

    # Apply max_connections limit
    if max_connections:
        selected = selected[:max_connections]

    return selected



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from dataclasses import dataclass
from logging import Logger
from typing import Any, List, Optional

import asyncclick as click
import asyncer
from jinja2 import Template
from pipelines.models.steps import CommandResult

ALL_RESULTS_KEY = "_run_all_results"

SUMMARY_TEMPLATE_STR = """
{% if command_results %}
Summary of commands results
========================
{% for command_result in command_results %}
{{ '✅' if  command_result.success else '❌' }} {{ command_result.command.name }}
{% endfor %}
{% endif %}
"""

DETAILS_TEMPLATE_STR = """
{% for command_result in command_results %}
{% if command_result.stdout or command_result.stderr %}
=================================

Details for {{ command_result.command.name }} command
{% if command_result.stdout %}
STDOUT:
{{ command_result.stdout }}
{% endif %}
{% if command_result.stderr %}
STDERR:
{{ command_result.stderr }}
{% endif %}
{% endif %}
{% endfor %}

"""


@dataclass
class LogOptions:
    quiet: bool = True
    help_message: Optional[str] = None


def log_command_results(
    ctx: click.Context,
    command_results: List[CommandResult],
    logger: Logger,
    options: LogOptions = LogOptions(),
) -> None:
    """
    Log the output of the subcommands run by `run_all_subcommands`.
    """

    if not options.quiet:
        details_template = Template(DETAILS_TEMPLATE_STR)
        details_message = details_template.render(command_results=command_results)
        logger.info(details_message)

    summary_template = Template(SUMMARY_TEMPLATE_STR)
    summary_message = summary_template.render(command_results=command_results)
    logger.info(summary_message)

    if options.help_message:
        logger.info(options.help_message)


async def invoke_commands_concurrently(
    ctx: click.Context, commands: List[click.Command]
) -> List[Any]:
    """
    Run click commands concurrently and return a list of their return values.
    """

    soon_command_executions_results = []
    async with asyncer.create_task_group() as command_task_group:
        for command in commands:
            soon_command_execution_result = command_task_group.soonify(command.invoke)(
                ctx
            )
            soon_command_executions_results.append(soon_command_execution_result)
    return [r.value for r in soon_command_executions_results]


async def invoke_commands_sequentially(
    ctx: click.Context, commands: List[click.Command]
) -> List[Any]:
    """
    Run click commands sequentially and return a list of their return values.
    """
    command_executions_results = []
    for command in commands:
        command_executions_results.append(await command.invoke(ctx))
    return command_executions_results


def get_all_sibling_commands(ctx: click.Context) -> List[click.Command]:
    """
    Get all sibling commands of the current command.
    """
    return [
        c for c in ctx.parent.command.commands.values() if c.name != ctx.command.name
    ]  # type: ignore



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

# Source: https://click.palletsprojects.com/en/8.1.x/complex/

import importlib
from typing import Any, Dict, List, Optional

import asyncclick as click


class LazyGroup(click.Group):
    """
    A click Group that can lazily load subcommands.
    """

    def __init__(
        self,
        *args: Any,
        lazy_subcommands: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        # lazy_subcommands is a map of the form:
        #
        #   {command-name} -> {module-name}.{command-object-name}
        #
        self.lazy_subcommands = lazy_subcommands or {}

    def list_commands(self, ctx: click.Context) -> List[str]:
        base = super().list_commands(ctx)
        lazy = sorted(self.lazy_subcommands.keys())
        return base + lazy

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _lazy_load(self, cmd_name: str) -> click.Command:
        # lazily loading a command, first get the module name and attribute name
        import_path = self.lazy_subcommands[cmd_name]
        modname, cmd_object_name = import_path.rsplit(".", 1)
        # do the import
        mod = importlib.import_module(modname)
        # get the Command object from that module
        cmd_object = getattr(mod, cmd_object_name)
        # check the result to make debugging easier
        if not isinstance(cmd_object, click.Command):
            print(f"{cmd_object} is of instance {type(cmd_object)}")
            raise ValueError(
                f"Lazy loading of {import_path} failed by returning "
                "a non-command object"
            )
        return cmd_object



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import logging
import os

import sentry_sdk

sentry_logger = logging.getLogger("sentry")


def setup_sentry():
    """
    Setup the sentry SDK if SENTRY_DSN is defined for the environment.

    Additionally TRACES_SAMPLE_RATE can be set 0-1 otherwise will default to 0.
    """
    from sentry_sdk.integrations.argv import ArgvIntegration
    from sentry_sdk.integrations.atexit import AtexitIntegration
    from sentry_sdk.integrations.dedupe import DedupeIntegration
    from sentry_sdk.integrations.logging import LoggingIntegration, ignore_logger
    from sentry_sdk.integrations.modules import ModulesIntegration
    from sentry_sdk.integrations.stdlib import StdlibIntegration

    # We ignore the Dagster internal logging to prevent a single error from being logged per node in the job graph
    ignore_logger("dagster")

    SENTRY_DSN = os.environ.get("SENTRY_DSN")
    SENTRY_ENVIRONMENT = os.environ.get("SENTRY_ENVIRONMENT")
    TRACES_SAMPLE_RATE = float(os.environ.get("SENTRY_TRACES_SAMPLE_RATE", 0))

    sentry_logger.info("Setting up Sentry with")
    sentry_logger.info(f"SENTRY_DSN: {SENTRY_DSN}")
    sentry_logger.info(f"SENTRY_ENVIRONMENT: {SENTRY_ENVIRONMENT}")
    sentry_logger.info(f"SENTRY_TRACES_SAMPLE_RATE: {TRACES_SAMPLE_RATE}")

    if SENTRY_DSN:
        sentry_sdk.init(
            dsn=SENTRY_DSN,
            traces_sample_rate=TRACES_SAMPLE_RATE,
            environment=SENTRY_ENVIRONMENT,
            default_integrations=False,
            integrations=[
                AtexitIntegration(),
                DedupeIntegration(),
                StdlibIntegration(),
                ModulesIntegration(),
                ArgvIntegration(),
                LoggingIntegration(),
            ],
        )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""CLI commands for connector registry operations.

This module provides CLI wrappers for registry operations. The core logic
lives in the `airbyte_ops_mcp.registry` capability module.

Command groups:
    airbyte-ops registry rc promote - Promote RC to stable
    airbyte-ops registry rc rollback - Rollback RC version override
    airbyte-ops registry connector list - List all connectors in registry
    airbyte-ops registry connector-version list - List versions for a connector
    airbyte-ops registry connector-version next - Compute next version tag (prerelease/RC)
    airbyte-ops registry connector-version yank - Mark a connector version as yanked
    airbyte-ops registry connector-version unyank - Remove yank marker from a connector version
    airbyte-ops registry connector-version metadata get - Read connector metadata from GCS
    airbyte-ops registry connector-version artifacts generate - Generate version artifacts locally via docker
    airbyte-ops registry connector-version artifacts publish - Publish version artifacts to GCS
    airbyte-ops registry store mirror --local|--gcs-bucket|--s3-bucket - Mirror entire registry for testing
    airbyte-ops registry store compile --store coral:dev|coral:prod - Compile registry indexes and sync latest/ dirs
    airbyte-ops registry marketing-stubs sync --store coral:prod - Sync connector_stubs.json to GCS
    airbyte-ops registry marketing-stubs check --store coral:prod - Compare local file with GCS
"""

from __future__ import annotations

import contextlib
import sys
from pathlib import Path
from typing import Annotated, Literal

import yaml
from cyclopts import App, Parameter

from airbyte_ops_mcp.cli._base import app
from airbyte_ops_mcp.cli._shared import (
    error_console,
    exit_with_error,
    print_error,
    print_json,
    print_success,
)
from airbyte_ops_mcp.github_api import (
    get_file_contents_at_ref,
    resolve_github_token,
)
from airbyte_ops_mcp.mcp.prerelease import (
    compute_prerelease_docker_image_tag,
)
from airbyte_ops_mcp.registry import (
    resolve_store_target,
)
from airbyte_ops_mcp.registry.compare import compare_stores
from airbyte_ops_mcp.registry.connector_stubs import (
    CONNECTOR_STUBS_FILE,
)
from airbyte_ops_mcp.registry.generate import generate_version_artifacts
from airbyte_ops_mcp.registry.registry_store_base import (
    RegistryStoreBase,
    get_registry_store,
)

# Create the registry sub-app
registry_app = App(
    name="registry", help="Connector registry operations (GCS metadata service)."
)
app.command(registry_app)

# Create the RC sub-app under registry
rc_app = App(name="rc", help="Release candidate lifecycle operations.")
registry_app.command(rc_app)

# Create the connector sub-app under registry
connector_app = App(name="connector", help="Connector listing operations.")
registry_app.command(connector_app)

# Create the connector-version sub-app under registry
connector_version_app = App(
    name="connector-version",
    help="Connector version operations (list, yank, unyank, artifacts).",
)
registry_app.command(connector_version_app)

# Create the metadata sub-app under connector-version
metadata_app = App(
    name="metadata",
    help="Connector metadata inspection.",
)
connector_version_app.command(metadata_app)

# Create the artifacts sub-app under connector-version
artifacts_app = App(
    name="artifacts",
    help="Version artifact generation and publishing.",
)
connector_version_app.command(artifacts_app)

# Create the store sub-app under registry (whole-registry operations)
store_app = App(
    name="store",
    help="Whole-registry store operations (mirror, compile).",
)
registry_app.command(store_app)

# Create the marketing-stubs sub-app under registry (for whole-file GCS operations)
marketing_stubs_app = App(
    name="marketing-stubs",
    help="Marketing connector stubs GCS operations (whole-file sync).",
)
registry_app.command(marketing_stubs_app)


AIRBYTE_REPO_OWNER = "airbytehq"
AIRBYTE_ENTERPRISE_REPO_NAME = "airbyte-enterprise"
AIRBYTE_REPO_NAME = "airbyte"
CONNECTOR_PATH_PREFIX = "airbyte-integrations/connectors"


def _resolve_store(
    store: str | None,
    connector_name: str | None = None,
) -> RegistryStoreBase:
    """Resolve `--store` to a `RegistryStoreBase` instance.

    Wraps `resolve_store_target` and `get_registry_store`, converting
    `ValueError` into a user-friendly CLI error via `exit_with_error`.
    """
    try:
        target = resolve_store_target(store=store, connector_name=connector_name)
    except ValueError as e:
        exit_with_error(str(e))
        raise  # unreachable; satisfies type checker
    return get_registry_store(target)


def _get_connector_version_from_github(
    connector_name: str,
    ref: str,
    token: str | None = None,
) -> str | None:
    """Fetch connector version from metadata.yaml via GitHub API.

    Args:
        connector_name: Connector name (e.g., "source-github")
        ref: Git ref (commit SHA, branch name, or tag)
        token: GitHub API token (optional for public repos)

    Returns:
        Version string from metadata.yaml, or None if not found.
    """
    path = f"{CONNECTOR_PATH_PREFIX}/{connector_name}/metadata.yaml"
    contents = get_file_contents_at_ref(
        owner=AIRBYTE_REPO_OWNER,
        repo=AIRBYTE_REPO_NAME,
        path=path,
        ref=ref,
        token=token,
    )
    if contents is None:
        return None

    metadata = yaml.safe_load(contents)
    return metadata.get("data", {}).get("dockerImageTag")


@connector_version_app.command(name="next")
def compute_next_version(
    name: Annotated[
        str,
        Parameter(help="Connector name (e.g., 'source-github')."),
    ],
    sha: Annotated[
        str,
        Parameter(help="Git commit SHA (full or at least 7 characters)."),
    ],
    base_version: Annotated[
        str | None,
        Parameter(
            help="Base version override. If not provided, fetched from metadata.yaml at the given SHA."
        ),
    ] = None,
) -> None:
    """Compute the next version tag for a connector.

    Outputs the version tag to stdout for easy capture in shell scripts.
    This is the single source of truth for pre-release version format.

    The command fetches the connector's metadata.yaml from GitHub at the given SHA
    to determine the base version. It also compares against the master branch and
    prints a warning to stderr if no version bump is detected.

    If --base-version is provided, it is used directly instead of fetching from GitHub.

    Examples:
        airbyte-ops registry connector-version next --name source-github --sha abcdef1234567
        # Output: 1.2.3-preview.abcdef1

        airbyte-ops registry connector-version next --name source-github --sha abcdef1234567 --base-version 1.2.3
        # Output: 1.2.3-preview.abcdef1 (uses provided version, skips GitHub API)
    """
    # Try to get a GitHub token (optional, but helps avoid rate limiting)
    # Token resolution may fail if no token is configured, which is fine for public repos
    token: str | None = None
    with contextlib.suppress(ValueError):
        token = resolve_github_token()

    # Determine base version
    version: str
    if base_version:
        version = base_version
    else:
        # Fetch version from metadata.yaml at the given SHA
        fetched_version = _get_connector_version_from_github(name, sha, token)
        if fetched_version is None:
            print(
                f"Error: Could not fetch metadata.yaml for {name} at ref {sha}",
                file=sys.stderr,
            )
            sys.exit(1)
        version = fetched_version

    # Compare with master branch version and warn if no bump detected
    master_version = _get_connector_version_from_github(name, "master", token)
    if master_version and master_version == version:
        print(
            f"Warning: No version bump detected for {name}. "
            f"Version {version} matches master branch.",
            file=sys.stderr,
        )

    # Compute and output the prerelease tag
    tag = compute_prerelease_docker_image_tag(version, sha)
    print(tag)


# =============================================================================
# RC LIFECYCLE COMMANDS
# =============================================================================


@rc_app.command(name="promote")
def rc_promote(
    name: Annotated[
        str,
        Parameter(help="Connector technical name (e.g., source-github)."),
    ],
    repo_path: Annotated[
        Path,
        Parameter(help="Path to the Airbyte monorepo. Defaults to current directory."),
    ] = Path.cwd(),
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be published without making changes."),
    ] = False,
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Auto-detected from connector name when omitted.",
        ),
    ] = None,
) -> None:
    """Promote a release candidate to stable.

    Copies the release candidate metadata to the 'latest' path in GCS,
    then deletes the release candidate metadata.

    Uses `--store` to select the registry store and environment.
    When omitted, the store is auto-detected from the connector name.

    Requires GCS_CREDENTIALS environment variable to be set.
    """
    if not repo_path.exists():
        exit_with_error(f"Repository path not found: {repo_path}")

    registry = _resolve_store(store, connector_name=name)

    result = registry.rc_promote(
        repo_path=repo_path,
        connector_name=name,
        dry_run=dry_run,
    )

    print_json(result.model_dump())

    if result.status == "failure":
        exit_with_error(result.message or "Operation failed", code=1)


@rc_app.command(name="rollback")
def rc_rollback(
    name: Annotated[
        str,
        Parameter(help="Connector technical name (e.g., source-github)."),
    ],
    repo_path: Annotated[
        Path,
        Parameter(help="Path to the Airbyte monorepo. Defaults to current directory."),
    ] = Path.cwd(),
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be published without making changes."),
    ] = False,
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Auto-detected from connector name when omitted.",
        ),
    ] = None,
) -> None:
    """Rollback a release candidate version override.

    Deletes both the release candidate metadata and the versioned metadata
    from GCS after verifying their hashes match.

    Uses `--store` to select the registry store and environment.
    When omitted, the store is auto-detected from the connector name.

    Requires GCS_CREDENTIALS environment variable to be set.
    """
    if not repo_path.exists():
        exit_with_error(f"Repository path not found: {repo_path}")

    registry = _resolve_store(store, connector_name=name)

    result = registry.rc_rollback(
        repo_path=repo_path,
        connector_name=name,
        dry_run=dry_run,
    )

    print_json(result.model_dump())

    if result.status == "failure":
        exit_with_error(result.message or "Operation failed", code=1)


# =============================================================================
# REGISTRY I/O - READ COMMANDS
# =============================================================================


@metadata_app.command(name="get")
def get_connector_version_metadata_cmd(
    name: Annotated[
        str,
        Parameter(
            help="Connector name (e.g., 'source-faker', 'destination-postgres')."
        ),
    ],
    version: Annotated[
        str,
        Parameter(help="Version to read (e.g., 'latest', '1.2.3')."),
    ] = "latest",
    format: Annotated[
        Literal["json", "raw"],
        Parameter(help="Output format: 'json' for JSON, 'raw' for YAML."),
    ] = "json",
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Auto-detected from connector name when omitted.",
        ),
    ] = None,
) -> None:
    """Read a connector version's metadata from the registry.

    Returns the full metadata.yaml content for a connector at the specified version.

    Requires GCS_CREDENTIALS environment variable to be set.

    Examples:
        airbyte-ops registry connector-version metadata get --name source-faker
        airbyte-ops registry connector-version metadata get --name source-faker --version 6.2.38
        airbyte-ops registry connector-version metadata get --name source-faker --store coral:prod
    """
    registry = _resolve_store(store, connector_name=name)

    try:
        metadata = registry.get_connector_metadata(
            connector_name=name,
            version=version,
        )
    except FileNotFoundError as e:
        exit_with_error(str(e), code=1)
    except Exception as e:
        exit_with_error(f"Error reading metadata: {e}", code=1)

    if format == "json":
        print_json(metadata)
    else:
        print(yaml.dump(metadata, default_flow_style=False))


@connector_app.command(name="list")
def list_connectors_cmd(
    format: Annotated[
        Literal["json", "text"],
        Parameter(
            help="Output format: 'json' for JSON array, 'text' for newline-separated."
        ),
    ] = "json",
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Defaults to 'coral:dev'.",
        ),
    ] = "coral:dev",
) -> None:
    """List all connectors in the registry.

    Scans the registry bucket to find all connectors that have metadata files.

    Requires GCS_CREDENTIALS environment variable to be set.
    """
    registry = _resolve_store(store)

    try:
        connectors = registry.list_connectors()
    except Exception as e:
        exit_with_error(f"Error listing connectors: {e}", code=1)

    if format == "json":
        print_json({"connectors": connectors, "count": len(connectors)})
    else:
        for connector in connectors:
            print(connector)


@connector_version_app.command(name="list")
def list_connector_versions_cmd(
    name: Annotated[
        str,
        Parameter(help="Connector name (e.g., 'source-faker')."),
    ],
    format: Annotated[
        Literal["json", "text"],
        Parameter(
            help="Output format: 'json' for JSON array, 'text' for newline-separated."
        ),
    ] = "json",
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Auto-detected from connector name when omitted.",
        ),
    ] = None,
) -> None:
    """List all versions of a connector in the registry.

    Scans the registry bucket to find all versions of a specific connector.

    Requires GCS_CREDENTIALS environment variable to be set.
    """
    registry = _resolve_store(store, connector_name=name)

    try:
        versions = registry.list_connector_versions(
            connector_name=name,
        )
    except Exception as e:
        exit_with_error(f"Error listing versions: {e}", code=1)

    if format == "json":
        print_json({"connector": name, "versions": versions, "count": len(versions)})
    else:
        for v in versions:
            print(v)


@marketing_stubs_app.command(name="check")
def marketing_stubs_check(
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Defaults to 'coral:dev'.",
        ),
    ] = "coral:dev",
    repo_root: Annotated[
        Path,
        Parameter(
            help="Path to the airbyte-enterprise repository root. Defaults to current directory."
        ),
    ] = Path.cwd(),
) -> None:
    """Compare local connector_stubs.json with the version in GCS.

    This command reads the entire local connector_stubs.json file and compares it
    with the version currently published in GCS.

    Exit codes:
        0: Local file matches GCS (check passed)
        1: Differences found (check failed)

    Output:
        STDOUT: JSON representation of the comparison result
        STDERR: Informational messages and comparison details

    Example:
        airbyte-ops registry marketing-stubs check --store coral:prod --repo-root /path/to/airbyte-enterprise
        airbyte-ops registry marketing-stubs check --store coral:dev
    """
    registry = _resolve_store(store)

    try:
        result = registry.marketing_stubs_check(repo_root=repo_root)
    except FileNotFoundError as e:
        exit_with_error(str(e))
    except ValueError as e:
        exit_with_error(str(e))

    error_console.print(
        f"Comparing local {CONNECTOR_STUBS_FILE} with {result.get('bucket', '')}/{result.get('path', '')}"
    )

    differences = result.get("differences", [])
    if differences:
        error_console.print(
            f"[yellow]Warning:[/yellow] {len(differences)} difference(s) found:"
        )
        for diff in differences:
            error_console.print(f"  {diff['id']}: {diff['status']}")
        print_json(result)
        sys.exit(1)

    error_console.print(
        f"[green]Local file is in sync with GCS ({result.get('local_count', 0)} stubs)[/green]"
    )
    print_json(result)


@marketing_stubs_app.command(name="sync")
def marketing_stubs_sync(
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Defaults to 'coral:dev'.",
        ),
    ] = "coral:dev",
    repo_root: Annotated[
        Path,
        Parameter(
            help="Path to the airbyte-enterprise repository root. Defaults to current directory."
        ),
    ] = Path.cwd(),
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be uploaded without making changes."),
    ] = False,
) -> None:
    """Sync local connector_stubs.json to GCS.

    This command uploads the entire local connector_stubs.json file to GCS,
    replacing the existing file. Use this after merging changes to master
    in the airbyte-enterprise repository.

    Exit codes:
        0: Sync successful (or dry-run completed)
        1: Error (file not found, validation failed, etc.)

    Output:
        STDOUT: JSON representation of the sync result
        STDERR: Informational messages and status updates

    Example:
        airbyte-ops registry marketing-stubs sync --store coral:prod --repo-root /path/to/airbyte-enterprise
        airbyte-ops registry marketing-stubs sync --store coral:dev
        airbyte-ops registry marketing-stubs sync --store coral:dev --dry-run
    """
    registry = _resolve_store(store)

    try:
        result = registry.marketing_stubs_sync(
            repo_root=repo_root,
            dry_run=dry_run,
        )
    except FileNotFoundError as e:
        exit_with_error(str(e))
    except ValueError as e:
        exit_with_error(str(e))

    bucket_name = result.get("bucket", "")
    path = result.get("path", "")
    stub_count = result.get("stub_count", 0)

    if dry_run:
        error_console.print(
            f"[DRY RUN] Would upload {stub_count} stubs to {bucket_name}/{path}"
        )
    else:
        error_console.print(
            f"[green]Synced {stub_count} stubs to {bucket_name}/{path}[/green]"
        )
    print_json(result)


# =============================================================================
# REGISTRY REBUILD COMMANDS
# =============================================================================


@store_app.command(name="mirror")
def mirror_cmd(
    local: Annotated[
        bool,
        Parameter(
            help="Write output to a local directory. Mutually exclusive with --gcs-bucket and --s3-bucket.",
            negative="",
        ),
    ] = False,
    gcs_bucket: Annotated[
        str | None,
        Parameter(
            help="Write output to a GCS bucket. Must not be the prod bucket. "
            "Mutually exclusive with --local and --s3-bucket.",
        ),
    ] = None,
    s3_bucket: Annotated[
        str | None,
        Parameter(
            help="Write output to an S3 bucket. "
            "Mutually exclusive with --local and --gcs-bucket.",
        ),
    ] = None,
    output_path_root: Annotated[
        str | None,
        Parameter(
            help="Root path/prefix for the output. For --local, this is a directory path "
            "(defaults to a new temp dir if omitted). For --gcs-bucket/--s3-bucket, "
            "this prefix is prepended to all blob paths.",
        ),
    ] = None,
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be rebuilt without writing any files."),
    ] = False,
    source_store: Annotated[
        str | None,
        Parameter(
            help="Source store to read from (e.g. 'coral:prod', 'coral:dev'). "
            "Defaults to 'coral:prod'.",
        ),
    ] = "coral:prod",
    connector_name: Annotated[
        tuple[str, ...] | None,
        Parameter(
            help="Only rebuild these connectors (by name). "
            "Can be specified multiple times, e.g. "
            "--connector-name source-faker --connector-name destination-bigquery.",
        ),
    ] = None,
) -> None:
    """Create a mirror of the connector registry from the source store.

    Reads all connector metadata from the source store and copies it
    to the specified output target. Supports local filesystem, GCS, and S3
    as output targets via fsspec.

    Output targets are mutually exclusive: specify exactly one of
    --local, --gcs-bucket, or --s3-bucket.

    The production bucket is categorically disallowed as an output target.

    Examples:
        airbyte-ops registry store mirror --local
        airbyte-ops registry store mirror --local --source-store coral:dev
        airbyte-ops registry store mirror --gcs-bucket dev-airbyte-cloud-connector-metadata-service-2 \
            --output-path-root test-run-123
        airbyte-ops registry store mirror --s3-bucket my-test-bucket --dry-run
    """
    # Validate mutually exclusive output targets
    targets = [local, gcs_bucket is not None, s3_bucket is not None]
    if sum(targets) != 1:
        exit_with_error(
            "Specify exactly one output target: --local, --gcs-bucket, or --s3-bucket."
        )

    if local:
        output_mode = "local"
    elif gcs_bucket is not None:
        output_mode = "gcs"
    else:
        output_mode = "s3"

    registry = _resolve_store(source_store)

    result = registry.mirror(
        output_mode=output_mode,
        output_path_root=output_path_root,
        gcs_bucket=gcs_bucket,
        s3_bucket=s3_bucket,
        dry_run=dry_run,
        connector_name=list(connector_name) if connector_name else None,
    )

    print_json(
        {
            "status": result.status,
            "source_bucket": result.source_bucket,
            "output_mode": result.output_mode,
            "output_root": result.output_root,
            "connectors_processed": result.connectors_processed,
            "blobs_copied": result.blobs_copied,
            "blobs_skipped": result.blobs_skipped,
            "error_count": len(result.errors),
            "dry_run": result.dry_run,
        }
    )

    if result.errors:
        for err in result.errors[:10]:
            print_error(err)
        if len(result.errors) > 10:
            print_error(f"... and {len(result.errors) - 10} more errors")

    if result.status == "success":
        print_success(result.summary())
    elif result.status == "dry-run":
        print_success(f"[DRY RUN] {result.summary()}")
    else:
        error_console.print(f"[yellow]{result.summary()}[/yellow]")


# =============================================================================
# VERSION YANK COMMANDS
# =============================================================================


@connector_version_app.command(name="yank")
def yank_cmd(
    name: Annotated[
        str,
        Parameter(help="Connector name (e.g., 'source-faker')."),
    ],
    version: Annotated[
        str,
        Parameter(help="Version to yank (e.g., '1.2.3')."),
    ],
    reason: Annotated[
        str,
        Parameter(help="Reason for yanking this version."),
    ] = "",
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be done without making changes."),
    ] = False,
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Auto-detected from connector name when omitted.",
        ),
    ] = None,
) -> None:
    """Mark a connector version as yanked.

    Writes a version-yank.yml marker file to the version's directory in GCS.
    Yanked versions are excluded when determining the latest version of a
    connector.

    Requires GCS_CREDENTIALS environment variable to be set.

    Examples:
        airbyte-ops registry connector-version yank --name source-faker --version 1.2.3
        airbyte-ops registry connector-version yank --name source-faker --version 1.2.3 --reason "Critical bug"
        airbyte-ops registry connector-version yank --name source-faker --version 1.2.3 --store coral:prod
    """
    registry = _resolve_store(store, connector_name=name)

    result = registry.yank(
        connector_name=name,
        version=version,
        reason=reason,
        dry_run=dry_run,
    )

    print_json(result.to_dict())

    if result.success:
        print_success(result.message)
    else:
        exit_with_error(result.message, code=1)


@connector_version_app.command(name="unyank")
def unyank_cmd(
    name: Annotated[
        str,
        Parameter(help="Connector name (e.g., 'source-faker')."),
    ],
    version: Annotated[
        str,
        Parameter(help="Version to unyank (e.g., '1.2.3')."),
    ],
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be done without making changes."),
    ] = False,
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod'). "
            "Auto-detected from connector name when omitted.",
        ),
    ] = None,
) -> None:
    """Remove the yank marker from a connector version.

    Deletes the version-yank.yml marker file from the version's directory in GCS,
    making the version eligible again when determining the latest version.

    Requires GCS_CREDENTIALS environment variable to be set.

    Examples:
        airbyte-ops registry connector-version unyank --name source-faker --version 1.2.3
        airbyte-ops registry connector-version unyank --name source-faker --version 1.2.3 --store coral:prod
    """
    registry = _resolve_store(store, connector_name=name)

    result = registry.unyank(
        connector_name=name,
        version=version,
        dry_run=dry_run,
    )

    print_json(result.to_dict())

    if result.success:
        print_success(result.message)
    else:
        exit_with_error(result.message, code=1)


# =============================================================================
# ARTIFACT GENERATION COMMANDS
# =============================================================================


@artifacts_app.command(name="generate")
def generate_version_artifacts_cmd(
    metadata_file: Annotated[
        Path,
        Parameter(help="Path to the connector's metadata.yaml file."),
    ],
    docker_image: Annotated[
        str,
        Parameter(
            help="Docker image to run spec against (e.g., 'airbyte/source-faker:6.2.38')."
        ),
    ],
    output_dir: Annotated[
        Path | None,
        Parameter(
            help="Directory to write artifacts to. If not specified, a temp directory is created."
        ),
    ] = None,
    repo_root: Annotated[
        Path | None,
        Parameter(
            help=(
                "Root of the Airbyte repo checkout (for resolving doc.md). "
                "If not specified, inferred by walking up from metadata-file."
            ),
        ),
    ] = None,
    dry_run: Annotated[
        bool,
        Parameter(
            help="Show what would be generated without running docker or writing files."
        ),
    ] = False,
    with_validate: Annotated[
        bool,
        Parameter(
            ("--with-validate/--no-validate",),
            help=(
                "Run metadata validators after generation (default: enabled). "
                "Use --no-validate to skip."
            ),
        ),
    ] = True,
) -> None:
    """Generate version artifacts for a connector locally.

    Runs the connector's docker image with DEPLOYMENT_MODE=cloud and
    DEPLOYMENT_MODE=oss to obtain both spec variants, then generates
    the registry entries (cloud.json, oss.json) by applying
    registryOverrides from the metadata.

    The generated metadata.yaml is enriched with git commit info, SBOM URL,
    and (when applicable) components SHA before writing.  Validation is run
    after generation by default; pass `--no-validate` to skip.

    This is a local-only operation -- no files are uploaded to GCS.
    Use `artifacts publish` to upload generated artifacts to GCS.

    Examples:
        airbyte-ops registry connector-version artifacts generate \\
            --metadata-file path/to/metadata.yaml \\
            --docker-image airbyte/source-faker:6.2.38

        airbyte-ops registry connector-version artifacts generate \\
            --metadata-file path/to/metadata.yaml \\
            --docker-image airbyte/source-faker:6.2.38 \\
            --output-dir ./artifacts --with-validate
    """
    result = generate_version_artifacts(
        metadata_file=metadata_file,
        docker_image=docker_image,
        output_dir=output_dir,
        repo_root=repo_root,
        dry_run=dry_run,
        with_validate=with_validate,
    )

    print_json(result.to_dict())

    if result.success:
        print_success(
            f"Generated {len(result.artifacts_written)} artifacts to {result.output_dir}"
        )
    else:
        all_errors = result.errors + result.validation_errors
        exit_with_error(
            f"Generation completed with {len(all_errors)} error(s): "
            + "; ".join(all_errors),
            code=1,
        )


@artifacts_app.command(name="publish")
def publish_version_artifacts_cmd(
    *,
    name: Annotated[
        str,
        Parameter(help="Connector name (e.g., 'source-faker')."),
    ],
    version: Annotated[
        str,
        Parameter(help="Version to publish artifacts for (e.g., '1.2.3')."),
    ],
    artifacts_dir: Annotated[
        Path,
        Parameter(
            help="Directory containing generated artifacts to publish (from 'artifacts generate')."
        ),
    ],
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod', 'coral:dev/prefix'). "
            "Auto-detected from connector name when omitted.",
        ),
    ] = None,
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be published without writing to GCS."),
    ] = False,
    with_validate: Annotated[
        bool,
        Parameter(
            ("--with-validate/--no-validate",),
            help=(
                "Validate metadata before uploading (default: enabled). "
                "Use --no-validate to skip."
            ),
        ),
    ] = True,
) -> None:
    """Publish version artifacts to GCS using fsspec rsync.

    Uploads locally generated artifacts (from `artifacts generate`) to the
    versioned path in GCS.  By default, metadata is validated before upload;
    pass `--no-validate` to skip.

    Uses `--store` to select the destination store and environment:

    * `coral:dev`              → coral dev bucket at root
    * `coral:prod`             → coral prod bucket at root
    * `coral:dev/aj-test100`   → coral dev bucket under `aj-test100/` prefix

    Requires GCS_CREDENTIALS environment variable to be set.

    Examples:
        airbyte-ops registry connector-version artifacts publish \\
            --name source-faker --version 6.2.38 \\
            --artifacts-dir ./artifacts --with-validate

        airbyte-ops registry connector-version artifacts publish \\
            --name source-faker --version 6.2.38 \\
            --artifacts-dir ./artifacts --store coral:prod

        airbyte-ops registry connector-version artifacts publish \\
            --name source-faker --version 6.2.38 \\
            --artifacts-dir ./artifacts --store coral:dev/aj-test100
    """
    registry = _resolve_store(store, connector_name=name)

    result = registry.publish_version_artifacts(
        connector_name=name,
        version=version,
        artifacts_dir=artifacts_dir,
        dry_run=dry_run,
        with_validate=with_validate,
    )

    print_json(
        {
            "status": result.status,
            "connector_name": result.connector_name,
            "version": result.version,
            "target": result.target,
            "gcs_destination": result.gcs_destination,
            "files_uploaded": result.files_uploaded,
            "errors": result.errors,
            "validation_errors": result.validation_errors,
            "dry_run": result.dry_run,
        }
    )

    if result.success:
        print_success(
            f"Published {len(result.files_uploaded)} artifacts for "
            f"{result.connector_name}@{result.version} → {result.gcs_destination}"
        )
    else:
        all_errors = result.errors + result.validation_errors
        exit_with_error(
            f"Publish completed with {len(all_errors)} error(s): "
            + "; ".join(all_errors),
            code=1,
        )


# =============================================================================
# COMPILE COMMAND
# =============================================================================


@store_app.command(name="compile")
def compile_cmd(
    store: Annotated[
        str | None,
        Parameter(
            help="Store target (e.g. 'coral:dev', 'coral:prod', 'coral:dev/prefix'). "
            "Defaults to 'coral:dev'.",
        ),
    ] = "coral:dev",
    connector_name: Annotated[
        tuple[str, ...] | None,
        Parameter(
            help="Only compile these connectors (can be repeated).",
        ),
    ] = None,
    dry_run: Annotated[
        bool,
        Parameter(help="Show what would be done without writing."),
    ] = False,
    with_secrets_mask: Annotated[
        bool,
        Parameter(
            help=(
                "Also regenerate specs_secrets_mask.yaml by scanning all "
                "connector specs for airbyte_secret properties."
            ),
        ),
    ] = False,
) -> None:
    """Compile the registry: sync latest/ dirs, write global and per-connector indexes.

    Scans all version directories in the target store, determines the latest GA
    semver per connector (excluding yanked and pre-release versions), ensures
    each `latest/` directory matches the computed latest, and writes:

    * `registries/v0/cloud_registry.json` -- global cloud registry index
    * `registries/v0/oss_registry.json`   -- global OSS registry index
    * `metadata/airbyte/<connector>/versions.json` -- per-connector version index

    With `--with-secrets-mask`, also regenerates:

    * `registries/v0/specs_secrets_mask.yaml` -- properties marked as secrets

    Uses efficient glob patterns for scanning (no file downloads during discovery).

    Requires GCS_CREDENTIALS environment variable to be set.

    Examples:
        airbyte-ops registry store compile --store coral:dev --dry-run

        airbyte-ops registry store compile --store coral:dev/aj-test100 \\
            --connector-name source-faker --connector-name destination-bigquery

        airbyte-ops registry store compile --store coral:prod --with-secrets-mask
    """
    registry = _resolve_store(store)

    result = registry.compile(
        connector_name=list(connector_name) if connector_name else None,
        dry_run=dry_run,
        with_secrets_mask=with_secrets_mask,
    )

    print_json(
        {
            "status": result.status,
            "target": result.target,
            "connectors_scanned": result.connectors_scanned,
            "versions_found": result.versions_found,
            "yanked_versions": result.yanked_versions,
            "latest_updated": result.latest_updated,
            "latest_already_current": result.latest_already_current,
            "cloud_registry_entries": result.cloud_registry_entries,
            "oss_registry_entries": result.oss_registry_entries,
            "version_indexes_written": result.version_indexes_written,
            "specs_secrets_mask_properties": result.specs_secrets_mask_properties,
            "errors": result.errors,
            "dry_run": result.dry_run,
        }
    )

    if result.status == "success" or result.status == "dry-run":
        print_success(result.summary())
    else:
        exit_with_error(
            f"Compile completed with {len(result.errors)} error(s): "
            + "; ".join(result.errors),
            code=1,
        )


# =============================================================================
# STORE COMPARE COMMAND
# =============================================================================


@store_app.command(name="compare")
def compare_cmd(
    store: Annotated[
        str,
        Parameter(
            help="Store target being evaluated (e.g. 'coral:dev/20260306-mirror-compile').",
        ),
    ],
    reference_store: Annotated[
        str,
        Parameter(
            help="Known-good reference store to compare against. Defaults to 'coral:prod'.",
        ),
    ] = "coral:prod",
    connector_name: Annotated[
        tuple[str, ...] | None,
        Parameter(
            help="Only compare these connectors (can be repeated).",
        ),
    ] = None,
    with_artifacts: Annotated[
        bool,
        Parameter(
            help="Compare per-connector artifact files "
            "(metadata.yaml, cloud.json, oss.json, spec.json).",
            negative="--no-artifacts",
        ),
    ] = True,
    with_indexes: Annotated[
        bool,
        Parameter(
            help="Compare global registry index files "
            "(cloud_registry.json, oss_registry.json).",
            negative="--no-indexes",
        ),
    ] = True,
) -> None:
    """Compare a store against a reference store and report differences.

    Evaluates the `--store` target against `--reference-store` (default:
    `coral:prod`) and reports per-connector artifact diffs and global index
    diffs.

    Requires GCS_CREDENTIALS environment variable to be set.

    Examples:
        airbyte-ops registry store compare --store coral:dev/20260306-mirror \\
            --reference-store coral:prod

        airbyte-ops registry store compare --store coral:dev/my-test \\
            --connector-name source-faker --no-indexes

        airbyte-ops registry store compare --store coral:dev/my-test \\
            --no-artifacts
    """
    store_target = _resolve_store(store)
    ref_target = _resolve_store(reference_store)

    store_prefix = f"{store_target.prefix}/" if store_target.prefix else ""
    ref_prefix = f"{ref_target.prefix}/" if ref_target.prefix else ""

    result = compare_stores(
        store_bucket=store_target.bucket_name,
        store_prefix=store_prefix,
        reference_bucket=ref_target.bucket_name,
        reference_prefix=ref_prefix,
        connector_name=list(connector_name) if connector_name else None,
        with_artifacts=with_artifacts,
        with_indexes=with_indexes,
    )

    print_json(result.to_dict())

    if result.status == "match":
        print_success(result.summary())
    elif result.status == "differences-found":
        error_console.print(f"[yellow]{result.summary()}[/yellow]")

        # Print a concise per-connector diff summary
        for diff in result.connector_diffs:
            if diff.status in ("only_in_store", "only_in_reference"):
                error_console.print(f"  {diff.connector}: {diff.status}")
            else:
                for ad in diff.artifact_diffs:
                    error_console.print(
                        f"  {diff.connector}/{ad.file}: {ad.status}"
                        + (f" ({ad.details})" if ad.details else "")
                    )

        for idx_diff in result.index_diffs:
            if idx_diff.status != "match":
                error_console.print(
                    f"  [index] {idx_diff.file}: {idx_diff.status}"
                    + (
                        f" (store={idx_diff.entry_count_store},"
                        f" ref={idx_diff.entry_count_reference})"
                        if idx_diff.entry_count_store or idx_diff.entry_count_reference
                        else ""
                    )
                )

        sys.exit(1)
    else:
        exit_with_error(
            f"Compare completed with {len(result.errors)} error(s): "
            + "; ".join(result.errors[:5]),
            code=1,
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Set

from consts import CIContext
from pipelines import main_logger
from pipelines.airbyte_ci.connectors.context import ConnectorContext
from pipelines.airbyte_ci.connectors.reports import ConnectorReport, Report
from pipelines.airbyte_ci.steps.pull_request import CreateOrUpdatePullRequest
from pipelines.helpers.git import get_modified_files
from pipelines.helpers.utils import transform_strs_to_paths

if TYPE_CHECKING:
    from anyio import Semaphore


## HELPER FUNCTIONS
async def get_connector_changes(context: ConnectorContext) -> Set[Path]:
    logger = main_logger
    all_modified_files = set(
        transform_strs_to_paths(
            await get_modified_files(
                context.git_branch,
                context.git_revision,
                context.diffed_branch,
                context.is_local,
                CIContext(context.ci_context),
                context.git_repo_url,
            )
        )
    )

    directory = context.connector.code_directory
    logger.info(f"Filtering to changes in {directory}")
    # get a list of files that are a child of this path
    connector_files = set(
        [file for file in all_modified_files if directory in file.parents]
    )
    # get doc too
    doc_path = context.connector.documentation_file_path

    if doc_path in all_modified_files:
        connector_files.add(doc_path)

    return connector_files


def replace_placeholder_with_pr_number(
    context: ConnectorContext, pr_number: int
) -> Set[Path]:
    current_doc = context.connector.documentation_file_path.read_text()
    updated_doc = current_doc.replace("*PR_NUMBER_PLACEHOLDER*", str(pr_number))
    context.connector.documentation_file_path.write_text(updated_doc)
    return {context.connector.documentation_file_path}


async def run_connector_pull_request_pipeline(
    context: ConnectorContext,
    semaphore: Semaphore,
    message: str,
    branch_id: str,
    title: str | None = None,
    body: str | None = None,
) -> Report | ConnectorReport | None:
    title = title or message
    body = body or ""
    async with semaphore, context:
        step_results = []
        modified_files = await get_connector_changes(context)

        create_or_update_pull_request = CreateOrUpdatePullRequest(context, skip_ci=True)

        if not modified_files:
            step_results.append(
                create_or_update_pull_request.skip(
                    "No changes detected in the connector directory."
                )
            )
            context.report = ConnectorReport(context, step_results, name="PULL REQUEST")
            return context.report

        create_or_update_pull_request_result = await create_or_update_pull_request.run(
            modified_files,
            branch_id,
            message,
            title,
            body,
        )
        step_results.append(create_or_update_pull_request_result)

        if not create_or_update_pull_request_result.success:
            return ConnectorReport(
                context,
                step_results,
                name="PULL REQUEST",
            )

        created_pr = create_or_update_pull_request_result.output
        modified_files.update(
            replace_placeholder_with_pr_number(context, created_pr.number)
        )
        update_pull_request = CreateOrUpdatePullRequest(context, skip_ci=False)
        update_pull_request_result = await update_pull_request.run(
            modified_files,
            branch_id,
            message,
            title,
            body,
        )
        step_results.append(update_pull_request_result)

        context.report = ConnectorReport(context, step_results, name="PULL REQUEST")
    return context.report



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Prod DB Access module for querying Airbyte Cloud Prod DB Replica.

This module provides:
- sql.py: SQL query templates and schema documentation
- db_engine.py: Database connection and engine management
- queries.py: Query execution functions
"""

from airbyte_ops_mcp.prod_db_access.db_engine import get_pool
from airbyte_ops_mcp.prod_db_access.sql import (
    SELECT_ACTORS_PINNED_TO_VERSION,
    SELECT_CONNECTIONS_BY_CONNECTOR,
    SELECT_CONNECTOR_VERSIONS,
    SELECT_DATAPLANES_LIST,
    SELECT_NEW_CONNECTOR_RELEASES,
    SELECT_ORG_WORKSPACES,
    SELECT_SUCCESSFUL_SYNCS_FOR_VERSION,
    SELECT_SYNC_RESULTS_FOR_VERSION,
    SELECT_WORKSPACE_INFO,
)

__all__ = [
    "SELECT_ACTORS_PINNED_TO_VERSION",
    "SELECT_CONNECTIONS_BY_CONNECTOR",
    "SELECT_CONNECTOR_VERSIONS",
    "SELECT_DATAPLANES_LIST",
    "SELECT_NEW_CONNECTOR_RELEASES",
    "SELECT_ORG_WORKSPACES",
    "SELECT_SUCCESSFUL_SYNCS_FOR_VERSION",
    "SELECT_SYNC_RESULTS_FOR_VERSION",
    "SELECT_WORKSPACE_INFO",
    "get_pool",
]



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Unit tests for the connector rollout monitoring API-based tool."""

from unittest.mock import MagicMock, patch

import pytest

from airbyte_ops_mcp.cloud_admin.api_client import get_actor_sync_info
from airbyte_ops_mcp.mcp.connector_rollout import (
    RolloutActorSelectionInfo,
    RolloutActorSyncStats,
    RolloutMonitoringResult,
    query_prod_rollout_monitoring_stats,
)


@pytest.mark.unit
@pytest.mark.parametrize(
    "api_response,expected_selection_info,expected_sync_stats",
    [
        pytest.param(
            {
                "data": {
                    "actor_selection_info": {
                        "num_actors": 100,
                        "num_pinned_to_connector_rollout": 10,
                        "num_actors_eligible_or_already_pinned": 50,
                    },
                    "syncs": {
                        "actor-1": {
                            "num_connections": 5,
                            "num_succeeded": 10,
                            "num_failed": 2,
                        },
                        "actor-2": {
                            "num_connections": 3,
                            "num_succeeded": 8,
                            "num_failed": 0,
                        },
                    },
                }
            },
            RolloutActorSelectionInfo(
                num_actors=100,
                num_pinned_to_connector_rollout=10,
                num_actors_eligible_or_already_pinned=50,
            ),
            [
                RolloutActorSyncStats(
                    actor_id="actor-1",
                    num_connections=5,
                    num_succeeded=10,
                    num_failed=2,
                ),
                RolloutActorSyncStats(
                    actor_id="actor-2", num_connections=3, num_succeeded=8, num_failed=0
                ),
            ],
            id="typical_response_with_multiple_actors",
        ),
        pytest.param(
            {
                "data": {
                    "actor_selection_info": {
                        "num_actors": 0,
                        "num_pinned_to_connector_rollout": 0,
                        "num_actors_eligible_or_already_pinned": 0,
                    },
                    "syncs": {},
                }
            },
            RolloutActorSelectionInfo(
                num_actors=0,
                num_pinned_to_connector_rollout=0,
                num_actors_eligible_or_already_pinned=0,
            ),
            [],
            id="empty_rollout_no_actors",
        ),
        pytest.param(
            {"data": {"actor_selection_info": {}, "syncs": {}}},
            RolloutActorSelectionInfo(
                num_actors=0,
                num_pinned_to_connector_rollout=0,
                num_actors_eligible_or_already_pinned=0,
            ),
            [],
            id="missing_fields_defaults_to_zero",
        ),
    ],
)
def test_query_prod_rollout_monitoring_stats(
    api_response: dict,
    expected_selection_info: RolloutActorSelectionInfo,
    expected_sync_stats: list[RolloutActorSyncStats],
) -> None:
    """Test query_prod_rollout_monitoring_stats transforms API response correctly."""
    mock_ctx = MagicMock()
    mock_ctx.request_context.lifespan_context = {}

    with patch(
        "airbyte_ops_mcp.mcp.connector_rollout._resolve_cloud_auth"
    ) as mock_resolve_auth, patch(
        "airbyte_ops_mcp.mcp.connector_rollout.api_client.get_actor_sync_info"
    ) as mock_get_sync_info:
        mock_resolve_auth.return_value = MagicMock(
            bearer_token="test-token", client_id=None, client_secret=None
        )
        mock_get_sync_info.return_value = api_response

        result = query_prod_rollout_monitoring_stats(
            rollout_id="test-rollout-id", ctx=mock_ctx
        )

        assert isinstance(result, RolloutMonitoringResult)
        assert result.rollout_id == "test-rollout-id"
        assert result.actor_selection_info == expected_selection_info

        assert len(result.actor_sync_stats) == len(expected_sync_stats)
        for actual, expected in zip(
            sorted(result.actor_sync_stats, key=lambda x: x.actor_id),
            sorted(expected_sync_stats, key=lambda x: x.actor_id),
        ):
            assert actual == expected


@pytest.mark.unit
def test_get_actor_sync_info_api_call() -> None:
    """Test get_actor_sync_info makes correct API call."""
    with patch(
        "airbyte_ops_mcp.cloud_admin.api_client._get_access_token"
    ) as mock_get_token, patch(
        "airbyte_ops_mcp.cloud_admin.api_client.requests.post"
    ) as mock_post:
        mock_get_token.return_value = "test-access-token"
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "data": {"actor_selection_info": {}, "syncs": {}}
        }
        mock_post.return_value = mock_response

        result = get_actor_sync_info(
            rollout_id="test-rollout-id",
            config_api_root="https://api.test.com",
            bearer_token="test-token",
        )

        mock_post.assert_called_once()
        call_args = mock_post.call_args
        assert (
            call_args[0][0]
            == "https://api.test.com/connector_rollout/get_actor_sync_info"
        )
        assert call_args[1]["json"] == {"id": "test-rollout-id"}
        assert "Authorization" in call_args[1]["headers"]
        assert result == {"data": {"actor_selection_info": {}, "syncs": {}}}


@pytest.mark.unit
def test_get_actor_sync_info_not_found() -> None:
    """Test get_actor_sync_info raises error when rollout not found."""
    from airbyte.exceptions import PyAirbyteInputError

    with patch(
        "airbyte_ops_mcp.cloud_admin.api_client._get_access_token"
    ) as mock_get_token, patch(
        "airbyte_ops_mcp.cloud_admin.api_client.requests.post"
    ) as mock_post:
        mock_get_token.return_value = "test-access-token"
        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_response.text = "Not found"
        mock_post.return_value = mock_response

        with pytest.raises(PyAirbyteInputError, match="Rollout not found"):
            get_actor_sync_info(
                rollout_id="nonexistent-rollout",
                config_api_root="https://api.test.com",
                bearer_token="test-token",
            )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Detect changed connectors in the Airbyte monorepo.

This module provides functionality to detect which connectors have been modified
by comparing git diffs between branches.
"""

from __future__ import annotations

import logging
import re
import subprocess
from dataclasses import dataclass
from enum import StrEnum
from functools import lru_cache
from pathlib import Path
from typing import Any

import yaml

CONNECTOR_PATH_PREFIX = "airbyte-integrations/connectors"
METADATA_FILE_NAME = "metadata.yaml"
GIT_DEFAULT_BRANCH = "origin/master"

logger = logging.getLogger(__name__)


class ConnectorLanguage(StrEnum):
    """Supported connector implementation languages."""

    PYTHON = "python"
    JAVA = "java"
    LOW_CODE = "low-code"
    MANIFEST_ONLY = "manifest-only"


class ConnectorType(StrEnum):
    """Connector types (source or destination)."""

    SOURCE = "source"
    DESTINATION = "destination"


class ConnectorSubtype(StrEnum):
    """Connector subtypes based on data source category."""

    API = "api"
    DATABASE = "database"
    FILE = "file"
    CUSTOM = "custom"


def get_modified_connectors(
    repo_path: str | Path,
    base_ref: str = GIT_DEFAULT_BRANCH,
    head_ref: str = "HEAD",
) -> list[str]:
    """Get list of connector IDs that have been modified.

    This function compares the git diff between base_ref and head_ref to determine
    which connectors have been modified. A connector is considered modified if any
    files within its directory have changed.

    Args:
        repo_path: Path to the Airbyte monorepo
        base_ref: Base git reference to compare against (default: "origin/master")
        head_ref: Head git reference to compare (default: "HEAD")

    Returns:
        List of connector technical names (e.g., ["source-faker", "destination-postgres"])

    Example:
        >>> connectors = get_changed_connectors("/path/to/airbyte", "origin/master")
        >>> print(connectors)
        ['source-faker', 'destination-postgres']
    """
    repo_path = Path(repo_path)

    # Get list of changed files using git diff
    try:
        result = subprocess.run(
            ["git", "diff", "--name-only", base_ref, head_ref],
            cwd=repo_path,
            capture_output=True,
            text=True,
            check=True,
            timeout=30,
        )
        changed_files = result.stdout.strip().split("\n")
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to get git diff: {e.stderr}") from e
    except subprocess.TimeoutExpired as e:
        raise RuntimeError("Git diff command timed out after 30 seconds") from e

    # Filter for connector files and extract connector names
    changed_connectors = set()
    for file_path in changed_files:
        if file_path.startswith(CONNECTOR_PATH_PREFIX + "/"):
            # Extract connector name from path
            # e.g., "airbyte-integrations/connectors/source-faker/..." -> "source-faker"
            path_parts = file_path.split("/")
            if len(path_parts) >= 3:
                connector_name = path_parts[2]
                changed_connectors.add(connector_name)

    return sorted(changed_connectors)


@lru_cache(maxsize=128)
def get_certified_connectors(repo_path: str | Path) -> set[str]:
    """Get set of all certified connector IDs from metadata.

    This function reads the metadata.yaml file for each connector to determine
    which connectors have "certified" support level.

    Args:
        repo_path: Path to the Airbyte monorepo

    Returns:
        Set of certified connector technical names

    Example:
        >>> certified = get_certified_connectors("/path/to/airbyte")
        >>> "source-postgres" in certified
        True
    """
    repo_path = Path(repo_path)
    connectors_dir = repo_path / CONNECTOR_PATH_PREFIX

    if not connectors_dir.exists():
        raise ValueError(f"Connectors directory not found: {connectors_dir}")

    certified_connectors = set()

    # Iterate through all connector directories
    for connector_dir in connectors_dir.iterdir():
        if not connector_dir.is_dir():
            continue

        metadata_file = connector_dir / METADATA_FILE_NAME
        if not metadata_file.exists():
            continue

        # Read metadata to check support level
        try:
            import yaml

            with open(metadata_file) as f:
                metadata = yaml.safe_load(f)

            support_level = metadata.get("data", {}).get("supportLevel")
            if support_level == "certified":
                certified_connectors.add(connector_dir.name)
        except Exception:
            # Skip connectors with invalid metadata
            continue

    return certified_connectors


def get_all_connectors(repo_path: str | Path) -> set[str]:
    """Get set of all connector IDs in the repository.

    Args:
        repo_path: Path to the Airbyte monorepo

    Returns:
        Set of all connector technical names

    Example:
        >>> all_connectors = get_all_connectors("/path/to/airbyte")
        >>> "source-faker" in all_connectors
        True
    """
    repo_path = Path(repo_path)
    connectors_dir = repo_path / CONNECTOR_PATH_PREFIX

    if not connectors_dir.exists():
        raise ValueError(f"Connectors directory not found: {connectors_dir}")

    return {p.name for p in connectors_dir.iterdir() if p.is_dir()}


def get_connector_metadata(
    repo_path: str | Path,
    connector_name: str,
) -> dict[str, Any] | None:
    """Get metadata for a specific connector.

    Args:
        repo_path: Path to the Airbyte monorepo
        connector_name: Technical name of the connector (e.g., "source-faker")

    Returns:
        The connector's metadata dict (the 'data' section), or None if not found

    Example:
        >>> metadata = get_connector_metadata("/path/to/airbyte", "source-faker")
        >>> metadata.get("supportLevel")
        'certified'
    """
    repo_path = Path(repo_path)
    connector_dir = repo_path / CONNECTOR_PATH_PREFIX / connector_name
    metadata_file = connector_dir / METADATA_FILE_NAME

    if not metadata_file.exists():
        return None

    try:
        with open(metadata_file) as f:
            metadata = yaml.safe_load(f)
        return metadata.get("data", {})
    except Exception:
        return None


def get_connectors_by_language(
    repo_path: str | Path,
    language: ConnectorLanguage,
) -> set[str]:
    """Get set of all connector IDs for a specific language.

    This function reads connector directories to determine which connectors
    use the specified language.

    Args:
        repo_path: Path to the Airbyte monorepo
        language: Language to filter by

    Returns:
        Set of connector technical names using the specified language

    Example:
        >>> python_connectors = get_connectors_by_language(
        ...     "/path/to/airbyte", ConnectorLanguage.PYTHON
        ... )
        >>> "source-faker" in python_connectors
        True
    """
    repo_path = Path(repo_path)
    connectors_dir = repo_path / CONNECTOR_PATH_PREFIX

    if not connectors_dir.exists():
        raise ValueError(f"Connectors directory not found: {connectors_dir}")

    language_value = language.value
    connectors_by_language = set()

    # Iterate through all connector directories
    for connector_dir in connectors_dir.iterdir():
        if not connector_dir.is_dir():
            continue

        # Determine language based on file structure
        connector_name = connector_dir.name
        detected_language = _detect_connector_language(connector_dir, connector_name)

        if detected_language == language_value:
            connectors_by_language.add(connector_dir.name)

    return connectors_by_language


def get_connectors_by_type(
    repo_path: str | Path,
    connector_type: ConnectorType,
) -> set[str]:
    """Get set of all connector IDs for a specific type (source or destination).

    This function reads connector directories to determine which connectors
    match the specified type based on their metadata.yaml or name prefix.

    Args:
        repo_path: Path to the Airbyte monorepo
        connector_type: Type to filter by (source or destination)

    Returns:
        Set of connector technical names matching the specified type

    Example:
        >>> source_connectors = get_connectors_by_type(
        ...     "/path/to/airbyte", ConnectorType.SOURCE
        ... )
        >>> "source-postgres" in source_connectors
        True
    """
    repo_path = Path(repo_path)
    connectors_dir = repo_path / CONNECTOR_PATH_PREFIX

    if not connectors_dir.exists():
        raise ValueError(f"Connectors directory not found: {connectors_dir}")

    type_value = connector_type.value
    connectors_by_type = set()

    for connector_dir in connectors_dir.iterdir():
        if not connector_dir.is_dir():
            continue

        connector_name = connector_dir.name

        # First try to get type from metadata.yaml
        metadata = get_connector_metadata(repo_path, connector_name)
        if metadata:
            metadata_type = metadata.get("connectorType")
            if metadata_type == type_value:
                connectors_by_type.add(connector_name)
                continue

        # Fallback to name prefix detection
        if connector_name.startswith(f"{type_value}-"):
            connectors_by_type.add(connector_name)

    return connectors_by_type


def get_connectors_by_subtype(
    repo_path: str | Path,
    connector_subtype: ConnectorSubtype,
) -> set[str]:
    """Get set of all connector IDs for a specific subtype (api, database, file, etc.).

    This function reads connector metadata.yaml files to determine which connectors
    match the specified subtype.

    Args:
        repo_path: Path to the Airbyte monorepo
        connector_subtype: Subtype to filter by (api, database, file, custom)

    Returns:
        Set of connector technical names matching the specified subtype

    Example:
        >>> database_connectors = get_connectors_by_subtype(
        ...     "/path/to/airbyte", ConnectorSubtype.DATABASE
        ... )
        >>> "source-postgres" in database_connectors
        True
    """
    repo_path = Path(repo_path)
    connectors_dir = repo_path / CONNECTOR_PATH_PREFIX

    if not connectors_dir.exists():
        raise ValueError(f"Connectors directory not found: {connectors_dir}")

    subtype_value = connector_subtype.value
    connectors_by_subtype = set()

    for connector_dir in connectors_dir.iterdir():
        if not connector_dir.is_dir():
            continue

        connector_name = connector_dir.name
        metadata = get_connector_metadata(repo_path, connector_name)

        if metadata:
            metadata_subtype = metadata.get("connectorSubtype")
            if metadata_subtype == subtype_value:
                connectors_by_subtype.add(connector_name)

    return connectors_by_subtype


@lru_cache(maxsize=1024)
def _detect_connector_language(connector_dir: Path, connector_name: str) -> str | None:
    """Detect the language of a connector based on its file structure.

    Args:
        connector_dir: Path to the connector directory
        connector_name: Technical name of the connector

    Returns:
        Language string (python, java, low-code, manifest-only) or None
    """
    # Check for manifest-only (manifest.yaml at root)
    if (connector_dir / "manifest.yaml").is_file():
        return "manifest-only"

    # Check for low-code (manifest.yaml in source directory)
    source_dir = connector_dir / connector_name.replace("-", "_")
    if (source_dir / "manifest.yaml").is_file():
        return "low-code"

    # Check for Python (setup.py or pyproject.toml)
    if (connector_dir / "setup.py").is_file() or (
        connector_dir / "pyproject.toml"
    ).is_file():
        return "python"

    # Check for Java/Kotlin (src/main/java or src/main/kotlin)
    if (connector_dir / "src" / "main" / "java").exists() or (
        connector_dir / "src" / "main" / "kotlin"
    ).exists():
        return "java"

    return None


@lru_cache(maxsize=128)
def get_connectors_with_local_cdk(repo_path: str | Path) -> set[str]:
    """Get set of connectors using local CDK reference instead of published version.

    This function detects connectors that are configured to use a local CDK
    checkout rather than a published CDK version. The detection method differs
    by connector language:
    - Python: Checks for path-based dependency in pyproject.toml
    - Java/Kotlin: Checks for useLocalCdk=true in build.gradle files

    Args:
        repo_path: Path to the Airbyte monorepo

    Returns:
        Set of connector technical names using local CDK reference

    Example:
        >>> local_cdk_connectors = get_connectors_with_local_cdk("/path/to/airbyte")
        >>> "destination-motherduck" in local_cdk_connectors
        True
    """
    repo_path = Path(repo_path)
    connectors_dir = repo_path / CONNECTOR_PATH_PREFIX

    if not connectors_dir.exists():
        raise ValueError(f"Connectors directory not found: {connectors_dir}")

    local_cdk_connectors = set()

    # Iterate through all connector directories
    for connector_dir in connectors_dir.iterdir():
        if not connector_dir.is_dir():
            continue

        connector_name = connector_dir.name
        detected_language = _detect_connector_language(connector_dir, connector_name)

        # Check for local CDK based on language
        if detected_language in ("python", "low-code") and _has_local_cdk_python(
            connector_dir
        ):
            # Python connectors: check pyproject.toml for path-based CDK dependency
            local_cdk_connectors.add(connector_name)
        elif detected_language == "java" and _has_local_cdk_java(connector_dir):
            # Java/Kotlin connectors: check build.gradle for useLocalCdk=true
            local_cdk_connectors.add(connector_name)

    return local_cdk_connectors


def _has_local_cdk_python(connector_dir: Path) -> bool:
    """Check if a Python connector uses local CDK reference.

    Args:
        connector_dir: Path to the connector directory

    Returns:
        True if connector uses local CDK reference
    """
    pyproject_file = connector_dir / "pyproject.toml"
    if not pyproject_file.exists():
        return False

    try:
        content = pyproject_file.read_text()
        # Look for path-based airbyte-cdk dependency
        # Pattern: airbyte-cdk = { path = "..." } or airbyte-cdk = {path = "..."}
        return bool(re.search(r"airbyte-cdk\s*=\s*\{[^}]*path\s*=", content))
    except Exception:
        return False


def _has_local_cdk_java(connector_dir: Path) -> bool:
    """Check if a Java/Kotlin connector uses local CDK reference.

    Mirrors the implementation from airbyte-ci/connectors/connector_ops/connector_ops/utils.py

    Args:
        connector_dir: Path to the connector directory

    Returns:
        True if connector uses local CDK reference
    """
    # Check both build.gradle and build.gradle.kts
    for build_file_name in ("build.gradle", "build.gradle.kts"):
        build_file = connector_dir / build_file_name
        if not build_file.exists():
            continue

        try:
            # Read file and strip inline comments
            contents = "\n".join(
                [
                    line.split("//")[0]  # Remove inline comments
                    for line in build_file.read_text().split("\n")
                ]
            )
            # Remove spaces and check for useLocalCdk=true
            contents = contents.replace(" ", "")
            if "useLocalCdk=true" in contents:
                return True
        except Exception:
            continue

    return False


@dataclass
class ConnectorListResult:
    """Result of listing connectors with filters."""

    connectors: list[str]
    count: int


def list_connectors(
    repo_path: str | Path,
    certified: bool | None = None,
    modified: bool | None = None,
    language_filter: set[str] | None = None,
    language_exclude: set[str] | None = None,
    connector_type: str | None = None,
    connector_subtype: str | None = None,
    base_ref: str | None = None,
    head_ref: str | None = None,
) -> ConnectorListResult:
    """List connectors in the Airbyte monorepo with flexible filtering.

    This is the core capability function that encapsulates all filtering logic.
    Both CLI and MCP layers should use this function.

    Args:
        repo_path: Path to the Airbyte monorepo
        certified: Filter by certification status (True=certified only,
            False=non-certified only, None=all)
        modified: Filter by modification status (True=modified only,
            False=not-modified only, None=all)
        language_filter: Set of languages to include (python, java, low-code, manifest-only)
        language_exclude: Set of languages to exclude (mutually exclusive with language_filter)
        connector_type: Filter by connector type (source, destination, or None for all)
        connector_subtype: Filter by connector subtype (api, database, file, custom, or None for all)
        base_ref: Base git reference for modification detection (default: "origin/main")
        head_ref: Head git reference for modification detection (default: "HEAD")

    Returns:
        ConnectorListResult with sorted list of connector names and count

    Raises:
        ValueError: If both language_filter and language_exclude are provided
    """
    # Validate mutual exclusivity of language filters
    if language_filter is not None and language_exclude is not None:
        raise ValueError(
            "Cannot specify both language_filter and language_exclude. "
            "Only one language filter parameter is accepted."
        )

    # Start with all connectors
    result = get_all_connectors(repo_path)

    # Apply certified filter
    if certified is not None:
        certified_set = get_certified_connectors(repo_path)
        if certified:
            # Include only certified
            result &= certified_set
        else:
            # Include only non-certified
            result -= certified_set

    # Apply modified filter
    if modified is not None:
        base = base_ref if base_ref is not None else GIT_DEFAULT_BRANCH
        head = head_ref if head_ref is not None else "HEAD"
        changed_set = set(get_modified_connectors(repo_path, base, head))
        if modified:
            # Include only modified
            result &= changed_set
        else:
            # Include only not-modified
            result -= changed_set

    # Apply language include filter
    if language_filter:
        # Get connectors for all specified languages and union them
        lang_result: set[str] = set()
        for lang in language_filter:
            lang_result |= get_connectors_by_language(
                repo_path, ConnectorLanguage(lang)
            )
        result &= lang_result

    # Apply language exclude filter
    if language_exclude:
        # Get connectors for all specified languages and exclude them
        for lang in language_exclude:
            excluded = get_connectors_by_language(repo_path, ConnectorLanguage(lang))
            result -= excluded

    # Apply connector type filter
    if connector_type is not None:
        type_set = get_connectors_by_type(repo_path, ConnectorType(connector_type))
        result &= type_set

    # Apply connector subtype filter
    if connector_subtype is not None:
        subtype_set = get_connectors_by_subtype(
            repo_path, ConnectorSubtype(connector_subtype)
        )
        result &= subtype_set

    return ConnectorListResult(
        connectors=sorted(result),
        count=len(result),
    )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Registry compile operations.

This module provides the core logic for the `registry compile` command,
which scans a registry bucket, determines the "latest" GA version for each
connector, ensures `latest/` directories are up-to-date, and writes
global registry JSON index files plus per-connector version indexes.

High-level algorithm
--------------------
1. **Scan** -- glob for `metadata.yaml` files to discover all
   (connector, version) tuples.  No file *contents* are downloaded.
2. **Yank set** -- glob for `version-yank.yml` markers to build an
   exclusion set.
3. **Compute latest** -- for each connector, pick the highest GA semver
   that is not yanked and not a pre-release.
4. **Fast-check latest/** -- glob for `version=*` marker files inside
   `latest/` directories.  Compare with the computed latest.  Only
   connectors whose marker disagrees (or is missing) need a resync.
5. **Resync stale latest/** -- `fsspec.generic.rsync` from the
   versioned directory to `latest/`, then write/update the
   `version=<x.y.z>` marker.
6. **Write index files**:
   a. Global `registries/v0/cloud_registry.json` and
      `registries/v0/oss_registry.json` (backwards-compatible).
   b. Per-connector `versions.json` (new).

When `with_secrets_mask=True`, the compile step also regenerates
`registries/v0/specs_secrets_mask.yaml` by scanning all connector specs
for properties marked `airbyte_secret: true`.
"""

from __future__ import annotations

import copy
import json
import logging
import os
import re
import sys
from dataclasses import dataclass, field
from typing import Any

import dpath.util
import gcsfs
import yaml
from packaging.version import InvalidVersion, Version

from airbyte_ops_mcp.registry._constants import (
    METADATA_FOLDER,
)

logger = logging.getLogger(__name__)

# Regex for the `version=<semver>` marker filename inside `latest/`
_VERSION_MARKER_RE = re.compile(r"^version=(.+)$")

# Registry index output path (relative to bucket root)
_REGISTRIES_PREFIX = "registries/v0"

# Specs secrets mask filename
_SPECS_SECRETS_MASK_FILENAME = "specs_secrets_mask.yaml"


# ---------------------------------------------------------------------------
# Result model
# ---------------------------------------------------------------------------


@dataclass
class CompileResult:
    """Result of a registry compile operation."""

    target: str
    connectors_scanned: int = 0
    versions_found: int = 0
    yanked_versions: int = 0
    latest_updated: int = 0
    latest_already_current: int = 0
    cloud_registry_entries: int = 0
    oss_registry_entries: int = 0
    version_indexes_written: int = 0
    specs_secrets_mask_properties: int = 0
    errors: list[str] = field(default_factory=list)
    dry_run: bool = False

    @property
    def status(self) -> str:
        if self.dry_run:
            return "dry-run"
        if self.errors:
            return "completed-with-errors"
        return "success"

    def summary(self) -> str:
        return (
            f"[{self.status}] Scanned {self.connectors_scanned} connectors, "
            f"{self.versions_found} versions ({self.yanked_versions} yanked). "
            f"Latest updated: {self.latest_updated}, "
            f"already current: {self.latest_already_current}. "
            f"Registry entries: cloud={self.cloud_registry_entries}, "
            f"oss={self.oss_registry_entries}. "
            f"Version indexes: {self.version_indexes_written}. "
            f"Specs secrets mask: {self.specs_secrets_mask_properties} properties. "
            f"Errors: {len(self.errors)}."
        )


@dataclass
class ConnectorVersionInfo:
    """Lightweight info about a single published version."""

    version: str
    yanked: bool = False
    is_latest: bool = False
    release_stage: str | None = None
    support_level: str | None = None


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _log_progress(msg: str, *args: object) -> None:
    """Log a progress message to both the logger and stderr."""
    logger.info(msg, *args)
    formatted = msg % args if args else msg
    print(formatted, file=sys.stderr, flush=True)


def _get_gcs_credentials_token() -> dict[str, str]:
    """Get GCS credentials as a token dict for gcsfs."""
    gcs_creds = os.environ.get("GCS_CREDENTIALS")
    if not gcs_creds:
        raise ValueError(
            "GCS_CREDENTIALS environment variable is required for GCS operations."
        )
    return json.loads(gcs_creds)


def _is_ga_version(version_str: str) -> bool:
    """Return True if the version string is a GA (non-prerelease) semver."""
    try:
        v = Version(version_str)
    except InvalidVersion:
        return False
    return not v.is_prerelease and not v.is_devrelease


def _parse_version(version_str: str) -> Version | None:
    """Parse a version string, returning None if it is not valid semver."""
    try:
        return Version(version_str)
    except InvalidVersion:
        return None


def _extract_connector_version(
    path: str, bucket: str, prefix: str = ""
) -> tuple[str, str] | None:
    """Extract (connector_name, version) from a GCS path.

    Expected path format:
        `<bucket>/[<prefix>/]metadata/airbyte/<connector>/<version>/metadata.yaml`

    Args:
        path: Full GCS path returned by glob.
        bucket: Bucket name.
        prefix: Optional path prefix (e.g. `"devin-test/"` or `""`).  Must
            include trailing `/` if non-empty.

    Returns None if the path does not match.
    """
    expected_prefix = f"{bucket}/{prefix}{METADATA_FOLDER}/airbyte/"
    if not path.startswith(expected_prefix):
        return None
    remainder = path[len(expected_prefix) :]
    parts = remainder.split("/")
    if len(parts) < 2:
        return None
    connector = parts[0]
    version = parts[1]
    # Skip non-version dirs like "latest", "release_candidate"
    if version in ("latest", "release_candidate"):
        return None
    return connector, version


# ---------------------------------------------------------------------------
# Step 1 + 2: Scan versions and yank markers via glob
# ---------------------------------------------------------------------------


def _scan_versions_and_yanks(
    fs: gcsfs.GCSFileSystem,
    bucket: str,
    prefix: str,
    connector_name: list[str] | None = None,
) -> tuple[dict[str, list[str]], set[tuple[str, str]]]:
    """Scan the registry for all (connector, version) tuples and yank markers.

    Uses efficient glob patterns -- no file contents are downloaded.

    Args:
        fs: Authenticated GCSFileSystem.
        bucket: Bucket name (with optional prefix already applied).
        prefix: Path prefix within the bucket (e.g. "" or "aj-test100/").
        connector_name: If provided, only scan these connector names.

    Returns:
        Tuple of:
        - dict mapping connector_name -> list of version strings
        - set of (connector_name, version) tuples that are yanked
    """
    base = f"{bucket}/{prefix}{METADATA_FOLDER}/airbyte"

    # Step 1: glob for metadata.yaml to discover all versions
    if connector_name:
        metadata_paths: list[str] = []
        for name in connector_name:
            pattern = f"{base}/{name}/*/metadata.yaml"
            metadata_paths.extend(fs.glob(pattern))
    else:
        pattern = f"{base}/*/*/metadata.yaml"
        metadata_paths = fs.glob(pattern)

    _log_progress("Glob found %d metadata.yaml files", len(metadata_paths))

    connector_versions: dict[str, list[str]] = {}
    for path in metadata_paths:
        parsed = _extract_connector_version(path, bucket, prefix)
        if parsed is None:
            continue
        connector, version = parsed
        connector_versions.setdefault(connector, []).append(version)

    # Step 2: glob for version-yank.yml markers
    if connector_name:
        yank_paths: list[str] = []
        for name in connector_name:
            yank_pattern = f"{base}/{name}/*/version-yank.yml"
            yank_paths.extend(fs.glob(yank_pattern))
    else:
        yank_pattern = f"{base}/*/*/version-yank.yml"
        yank_paths = fs.glob(yank_pattern)

    _log_progress("Glob found %d version-yank.yml markers", len(yank_paths))

    yanked: set[tuple[str, str]] = set()
    for path in yank_paths:
        # Path: bucket/[prefix/]metadata/airbyte/<connector>/<version>/version-yank.yml
        parts = path.split("/")
        # Find "metadata" index, then airbyte/<connector>/<version>
        try:
            meta_idx = parts.index("metadata")
            connector = parts[meta_idx + 2]
            version = parts[meta_idx + 3]
            yanked.add((connector, version))
        except (ValueError, IndexError):
            logger.warning("Could not parse yank path: %s", path)

    return connector_versions, yanked


# ---------------------------------------------------------------------------
# Step 3: Compute latest GA version per connector
# ---------------------------------------------------------------------------


def _compute_latest_versions(
    connector_versions: dict[str, list[str]],
    yanked: set[tuple[str, str]],
) -> dict[str, str]:
    """For each connector, determine the highest GA semver that is not yanked.

    Returns:
        dict mapping connector_name -> latest version string.
        Connectors with no eligible GA versions are omitted.
    """
    latest: dict[str, str] = {}
    for connector, versions in connector_versions.items():
        candidates: list[tuple[Version, str]] = []
        for v_str in versions:
            if (connector, v_str) in yanked:
                continue
            if not _is_ga_version(v_str):
                continue
            parsed = _parse_version(v_str)
            if parsed is not None:
                candidates.append((parsed, v_str))
        if candidates:
            candidates.sort(reverse=True)
            latest[connector] = candidates[0][1]
        else:
            logger.warning(
                "No eligible GA version found for %s (%d total versions, all yanked or pre-release).",
                connector,
                len(versions),
            )
    return latest


# ---------------------------------------------------------------------------
# Step 4: Check version markers in latest/ directories
# ---------------------------------------------------------------------------


def _scan_latest_markers(
    fs: gcsfs.GCSFileSystem,
    bucket: str,
    prefix: str,
    connector_name: list[str] | None = None,
) -> dict[str, str]:
    """Glob for `version=*` marker files in `latest/` directories.

    Returns:
        dict mapping connector_name -> current marker version string.
    """
    base = f"{bucket}/{prefix}{METADATA_FOLDER}/airbyte"

    if connector_name:
        marker_paths: list[str] = []
        for name in connector_name:
            marker_pattern = f"{base}/{name}/latest/version=*"
            marker_paths.extend(fs.glob(marker_pattern))
    else:
        marker_pattern = f"{base}/*/latest/version=*"
        marker_paths = fs.glob(marker_pattern)

    _log_progress("Glob found %d version marker files in latest/", len(marker_paths))

    markers: dict[str, str] = {}
    for path in marker_paths:
        filename = path.split("/")[-1]
        match = _VERSION_MARKER_RE.match(filename)
        if not match:
            continue
        version_str = match.group(1)
        # Extract connector name from path
        parts = path.split("/")
        try:
            meta_idx = parts.index("metadata")
            connector = parts[meta_idx + 2]
            markers[connector] = version_str
        except (ValueError, IndexError):
            logger.warning("Could not parse marker path: %s", path)

    return markers


# ---------------------------------------------------------------------------
# Step 5: Resync stale latest/ directories
# ---------------------------------------------------------------------------


def _sync_latest_dir(
    fs: gcsfs.GCSFileSystem,
    bucket: str,
    prefix: str,
    connector: str,
    version: str,
    old_marker_version: str | None,
    dry_run: bool = False,
) -> None:
    """Copy files from the versioned dir to `latest/` and update the marker.

    Uses the already-authenticated `fs` instance to copy files individually,
    avoiding credential-propagation issues with `fsspec.generic.rsync`.
    """
    base = f"{bucket}/{prefix}{METADATA_FOLDER}/airbyte/{connector}"
    source_dir = f"{base}/{version}"
    dest_dir = f"{base}/latest"

    if dry_run:
        _log_progress(
            "  [DRY RUN] Would sync %s/ → %s/ and write marker version=%s",
            source_dir,
            dest_dir,
            version,
        )
        return

    _log_progress("  Syncing %s/ → %s/", source_dir, dest_dir)

    # List source files and copy each to latest/
    source_files = fs.ls(source_dir, detail=False)
    source_filenames = {src_path.split("/")[-1] for src_path in source_files}

    for src_path in source_files:
        filename = src_path.split("/")[-1]
        dst_path = f"{dest_dir}/{filename}"
        fs.copy(src_path, dst_path)

    # Delete stale files from latest/ that no longer exist in the source version
    try:
        dest_files = fs.ls(dest_dir, detail=False)
    except FileNotFoundError:
        dest_files = []

    for dst_path in dest_files:
        filename = dst_path.split("/")[-1]
        # Preserve version marker files — they are managed separately below
        if filename.startswith("version="):
            continue
        if filename not in source_filenames:
            try:
                fs.rm(dst_path)
                _log_progress("    Deleted stale file from latest/: %s", filename)
            except Exception as exc:
                logger.warning("Failed to delete stale file %s: %s", dst_path, exc)

    # Delete old marker if present and different
    if old_marker_version and old_marker_version != version:
        old_marker_path = f"{base}/latest/version={old_marker_version}"
        try:
            if fs.exists(old_marker_path):
                fs.rm(old_marker_path)
        except Exception as exc:
            logger.warning("Failed to delete old marker %s: %s", old_marker_path, exc)

    # Write new marker (zero-byte file)
    marker_path = f"{base}/latest/version={version}"
    with fs.open(marker_path, "wb") as f:
        f.write(b"")

    _log_progress("  Wrote marker: version=%s", version)


def _apply_overrides_to_latest_entry(
    fs: gcsfs.GCSFileSystem,
    bucket: str,
    prefix: str,
    connector: str,
    version: str,
) -> None:
    """Post-process `latest/` registry entries after syncing from a versioned dir.

    When the compile step copies files from `{version}/` to `latest/`, the
    embedded `cloud.json` and `oss.json` still contain version-specific
    values.  This function:

    1. Applies `registryOverrides` from the metadata so that the `latest/`
       entry reflects the overridden values (e.g. a pinned `dockerImageTag`).
    2. When an override pins `dockerImageTag` to a different version, copies
       version-sensitive fields (`spec`, `packageInfo`) from the pinned
       version's entry so that `latest/` is fully consistent.
    3. Updates `generated.source_file_info.metadata_file_path` to reference
       the `latest/` path instead of the versioned path.
    """
    base = f"{bucket}/{prefix}{METADATA_FOLDER}/airbyte/{connector}"
    latest_dir = f"{base}/latest"

    # Read metadata.yaml from latest/ to get registry overrides.
    metadata_path = f"{latest_dir}/metadata.yaml"
    try:
        with fs.open(metadata_path, "r") as f:
            raw_metadata = yaml.safe_load(f)
    except FileNotFoundError:
        logger.warning(
            "No metadata.yaml in latest/ for %s, skipping override application",
            connector,
        )
        return
    except Exception as exc:
        logger.warning("Failed to read metadata for %s: %s", connector, exc)
        return

    metadata_data: dict[str, Any] = raw_metadata.get("data", {})
    registry_overrides = metadata_data.get("registryOverrides", {})

    latest_metadata_file_path = (
        f"{METADATA_FOLDER}/airbyte/{connector}/latest/metadata.yaml"
    )

    for registry_type in ("cloud", "oss"):
        entry_path = f"{latest_dir}/{registry_type}.json"
        try:
            if not fs.exists(entry_path):
                continue
            with fs.open(entry_path, "r") as f:
                entry = json.load(f)
        except Exception as exc:
            logger.warning(
                "Failed to read %s.json for %s: %s", registry_type, connector, exc
            )
            continue

        modified = False

        # --- Apply registry overrides (skip_docker_image_tag=False for latest) ---
        overrides = copy.deepcopy(registry_overrides.get(registry_type, {}))
        overrides.pop("enabled", None)
        overrides = {k: v for k, v in overrides.items() if v is not None}

        if overrides:
            entry.update(overrides)
            modified = True

        # --- When dockerImageTag is overridden to a different version, copy
        #     version-sensitive fields from the pinned version's entry. ---
        pinned_tag = overrides.get("dockerImageTag")
        if pinned_tag and pinned_tag != version:
            pinned_entry_path = f"{base}/{pinned_tag}/{registry_type}.json"
            try:
                if fs.exists(pinned_entry_path):
                    with fs.open(pinned_entry_path, "r") as fp:
                        pinned_entry = json.load(fp)
                    for field_name in ("spec", "packageInfo"):
                        if field_name in pinned_entry:
                            entry[field_name] = pinned_entry[field_name]
                            modified = True
                else:
                    logger.warning(
                        "Pinned version %s/%s.json not found for %s",
                        pinned_tag,
                        registry_type,
                        connector,
                    )
            except Exception as exc:
                logger.warning(
                    "Failed to read pinned entry %s/%s.json for %s: %s",
                    pinned_tag,
                    registry_type,
                    connector,
                    exc,
                )

        # --- Update metadata_file_path to point to latest/ ---
        generated = entry.get("generated")
        if isinstance(generated, dict):
            source_file_info = generated.get("source_file_info")
            if isinstance(source_file_info, dict):
                current_path = source_file_info.get("metadata_file_path")
                if current_path != latest_metadata_file_path:
                    source_file_info["metadata_file_path"] = latest_metadata_file_path
                    modified = True

        if modified:
            content = json.dumps(entry, indent=2, sort_keys=True) + "\n"
            with fs.open(entry_path, "w") as fout:
                fout.write(content)


# ---------------------------------------------------------------------------
# Step 6a: Compile global registry JSONs
# ---------------------------------------------------------------------------


def _compile_global_registry(
    fs: gcsfs.GCSFileSystem,
    bucket: str,
    prefix: str,
    latest_versions: dict[str, str],
    registry_type: str,
) -> list[dict[str, Any]]:
    """Read cloud.json or oss.json from each connector's latest/ dir.

    Args:
        registry_type: "cloud" or "oss".

    Returns:
        List of registry entry dicts.
    """
    base = f"{bucket}/{prefix}{METADATA_FOLDER}/airbyte"
    entries: list[dict[str, Any]] = []

    for connector in sorted(latest_versions):
        json_path = f"{base}/{connector}/latest/{registry_type}.json"
        try:
            if not fs.exists(json_path):
                continue
            with fs.open(json_path, "r") as f:
                entry = json.load(f)
            entries.append(entry)
        except Exception as exc:
            logger.warning(
                "Failed to read %s for %s: %s", registry_type, connector, exc
            )

    return entries


def _build_global_registry_json(entries: list[dict[str, Any]]) -> dict[str, Any]:
    """Assemble the global registry JSON from a list of connector entries.

    Splits entries into sources and destinations based on the presence of
    `sourceDefinitionId` or `destinationDefinitionId`.
    """
    sources = []
    destinations = []
    for entry in entries:
        if "sourceDefinitionId" in entry:
            sources.append(entry)
        elif "destinationDefinitionId" in entry:
            destinations.append(entry)
        else:
            # Fallback: guess from dockerRepository
            docker_repo = entry.get("dockerRepository", "")
            if "destination" in docker_repo:
                destinations.append(entry)
            else:
                sources.append(entry)

    return {"sources": sources, "destinations": destinations}


# ---------------------------------------------------------------------------
# Step 6c: Specs secrets mask
# ---------------------------------------------------------------------------


def _extract_secret_property_names(
    entries: list[dict[str, Any]],
) -> set[str]:
    """Extract property names marked as `airbyte_secret` from connector specs.

    Walks the `spec.connectionSpecification.properties` tree of each entry
    and collects the leaf property name for every node that has
    `airbyte_secret: true`.

    Args:
        entries: List of registry entry dicts (from cloud or oss registries).

    Returns:
        A set of property names that are marked as secrets.
    """
    secret_names: set[str] = set()
    for entry in entries:
        spec = entry.get("spec", {})
        conn_spec = spec.get("connectionSpecification", {})
        properties = conn_spec.get("properties")
        if not properties:
            continue
        for type_path, _ in dpath.util.search(properties, "**/type", yielded=True):
            absolute_path = f"/{type_path}"
            if "/" in type_path:
                property_path, _ = absolute_path.rsplit(sep="/", maxsplit=1)
            else:
                property_path = absolute_path
            try:
                property_definition = dpath.util.get(properties, property_path)
            except KeyError:
                continue
            if isinstance(property_definition, dict) and property_definition.get(
                "airbyte_secret", False
            ):
                secret_names.add(property_path.split("/")[-1])
    return secret_names


# ---------------------------------------------------------------------------
# Step 6b: Per-connector version index
# ---------------------------------------------------------------------------


def _build_version_index(
    connector: str,
    versions: list[str],
    yanked: set[tuple[str, str]],
    latest_version: str | None,
    fs: gcsfs.GCSFileSystem,
    bucket: str,
    prefix: str,
) -> dict[str, Any]:
    """Build the per-connector versions.json content.

    For each version, reads `metadata.yaml` to extract `releaseStage`
    and `supportLevel` (only for the latest version to keep scanning fast).
    """
    base = f"{bucket}/{prefix}{METADATA_FOLDER}/airbyte/{connector}"

    version_entries: list[dict[str, Any]] = []
    for v_str in sorted(
        versions, key=lambda s: _parse_version(s) or Version("0"), reverse=True
    ):
        is_yanked = (connector, v_str) in yanked
        is_latest = v_str == latest_version

        entry: dict[str, Any] = {
            "version": v_str,
            "yanked": is_yanked,
            "is_latest": is_latest,
        }

        # For the latest version, enrich with metadata fields
        if is_latest:
            try:
                meta_path = f"{base}/{v_str}/metadata.yaml"
                with fs.open(meta_path, "r") as f:
                    raw = yaml.safe_load(f)
                data = raw.get("data", {})
                entry["release_stage"] = data.get("releaseStage")
                entry["support_level"] = data.get("supportLevel")
            except Exception as exc:
                logger.warning(
                    "Failed to read metadata for %s@%s: %s", connector, v_str, exc
                )

        version_entries.append(entry)

    definition_id = None
    if latest_version:
        # Try to get the definition ID from the latest version's metadata
        try:
            meta_path = f"{base}/{latest_version}/metadata.yaml"
            with fs.open(meta_path, "r") as f:
                raw = yaml.safe_load(f)
            data = raw.get("data", {})
            definition_id = data.get("definitionId")
        except Exception:
            pass

    result: dict[str, Any] = {
        "connector": connector,
        "versions": version_entries,
    }
    if definition_id:
        result["definition_id"] = definition_id

    return result


# ---------------------------------------------------------------------------
# Public API
# ---------------------------------------------------------------------------


def compile_registry(
    bucket_name: str,
    prefix: str = "",
    connector_name: list[str] | None = None,
    dry_run: bool = False,
    with_secrets_mask: bool = False,
) -> CompileResult:
    """Compile the registry: sync latest/ dirs and write index files.

    Steps:
        1. Glob for all `metadata.yaml` to discover (connector, version) pairs.
        2. Glob for `version-yank.yml` to build the yanked set.
        3. Compute the latest GA semver per connector.
        4. Glob for `version=*` markers in `latest/` dirs for a fast check.
        5. Rsync stale `latest/` dirs from the calculated version.
        6. Write global registry JSONs and per-connector `versions.json`.
        6c. (Optional) Regenerate `specs_secrets_mask.yaml`.

    Args:
        bucket_name: GCS bucket name to compile from/to.
        prefix: Optional path prefix within the bucket.
        connector_name: If provided, only compile these connectors.
        dry_run: If True, report what would be done without writing.
        with_secrets_mask: If True, regenerate `specs_secrets_mask.yaml`.

    Returns:
        A `CompileResult` describing what was done.
    """
    prefix = f"{prefix}/" if prefix else ""

    target_label = f"{bucket_name}/{prefix}" if prefix else bucket_name
    result = CompileResult(target=target_label, dry_run=dry_run)

    token = _get_gcs_credentials_token()
    fs = gcsfs.GCSFileSystem(token=token)

    # --- Step 1 + 2: Scan versions and yanks ---
    _log_progress("Step 1-2: Scanning versions and yank markers...")
    connector_versions, yanked = _scan_versions_and_yanks(
        fs, bucket_name, prefix, connector_name
    )
    result.connectors_scanned = len(connector_versions)
    result.versions_found = sum(len(v) for v in connector_versions.values())
    result.yanked_versions = len(yanked)
    _log_progress(
        "  Found %d connectors, %d versions, %d yanked",
        result.connectors_scanned,
        result.versions_found,
        result.yanked_versions,
    )

    # --- Step 3: Compute latest ---
    _log_progress("Step 3: Computing latest GA version per connector...")
    latest_versions = _compute_latest_versions(connector_versions, yanked)
    _log_progress("  Computed latest for %d connectors", len(latest_versions))

    # --- Step 4: Check existing latest markers ---
    _log_progress("Step 4: Checking existing latest/ markers...")
    existing_markers = _scan_latest_markers(fs, bucket_name, prefix, connector_name)
    _log_progress("  Found %d existing markers", len(existing_markers))

    stale_connectors: list[str] = []
    for connector, expected_version in latest_versions.items():
        current_marker = existing_markers.get(connector)
        if current_marker == expected_version:
            result.latest_already_current += 1
        else:
            stale_connectors.append(connector)

    _log_progress(
        "  %d connectors need latest/ update, %d already current",
        len(stale_connectors),
        result.latest_already_current,
    )

    # --- Step 5: Resync stale latest/ dirs ---
    if stale_connectors:
        _log_progress(
            "Step 5: Syncing %d stale latest/ directories...", len(stale_connectors)
        )
        for connector in sorted(stale_connectors):
            version = latest_versions[connector]
            old_marker = existing_markers.get(connector)
            try:
                _sync_latest_dir(
                    fs, bucket_name, prefix, connector, version, old_marker, dry_run
                )
                if not dry_run:
                    _apply_overrides_to_latest_entry(
                        fs, bucket_name, prefix, connector, version
                    )
                result.latest_updated += 1
            except Exception as exc:
                error_msg = f"Failed to sync latest/ for {connector}: {exc}"
                logger.error(error_msg)
                result.errors.append(error_msg)
                # Remove the version marker so the next compile retries this
                # connector instead of treating it as "already current".
                marker_path = (
                    f"{bucket_name}/{prefix}{METADATA_FOLDER}/airbyte"
                    f"/{connector}/latest/version={version}"
                )
                try:
                    if fs.exists(marker_path):
                        fs.rm(marker_path)
                        logger.info(
                            "Deleted stale marker %s after failure", marker_path
                        )
                except Exception as marker_exc:
                    logger.warning(
                        "Could not delete stale marker %s: %s",
                        marker_path,
                        marker_exc,
                    )
    else:
        _log_progress("Step 5: All latest/ directories are current, nothing to sync.")

    # --- Step 6a: Compile global registry JSONs ---
    _log_progress("Step 6a: Compiling global registry JSON files...")
    all_registry_entries: list[dict[str, Any]] = []  # collected for Step 6c
    for registry_type in ("cloud", "oss"):
        entries = _compile_global_registry(
            fs, bucket_name, prefix, latest_versions, registry_type
        )
        all_registry_entries.extend(entries)
        registry_json = _build_global_registry_json(entries)

        output_path = (
            f"{bucket_name}/{prefix}{_REGISTRIES_PREFIX}/{registry_type}_registry.json"
        )
        entry_count = len(registry_json["sources"]) + len(registry_json["destinations"])

        if registry_type == "cloud":
            result.cloud_registry_entries = entry_count
        else:
            result.oss_registry_entries = entry_count

        if dry_run:
            _log_progress(
                "  [DRY RUN] Would write %s_registry.json (%d entries)",
                registry_type,
                entry_count,
            )
        else:
            content = json.dumps(registry_json, indent=2, sort_keys=True) + "\n"
            with fs.open(output_path, "w") as f:
                f.write(content)
            _log_progress(
                "  Wrote %s_registry.json (%d entries)",
                registry_type,
                entry_count,
            )

    # --- Step 6b: Per-connector version indexes ---
    _log_progress("Step 6b: Writing per-connector version indexes...")
    base = f"{bucket_name}/{prefix}{METADATA_FOLDER}/airbyte"
    for connector in sorted(connector_versions):
        versions = connector_versions[connector]
        latest_v = latest_versions.get(connector)
        index = _build_version_index(
            connector, versions, yanked, latest_v, fs, bucket_name, prefix
        )
        index_path = f"{base}/{connector}/versions.json"

        if dry_run:
            _log_progress(
                "  [DRY RUN] Would write %s/versions.json (%d versions)",
                connector,
                len(versions),
            )
        else:
            content = json.dumps(index, indent=2, sort_keys=True) + "\n"
            with fs.open(index_path, "w") as f:
                f.write(content)

        result.version_indexes_written += 1

    # --- Step 6c: Specs secrets mask (optional) ---
    if with_secrets_mask:
        if connector_name:
            _log_progress(
                "Step 6c: SKIPPED — --with-secrets-mask requires all connectors "
                "but --connector-name is set. Run without --connector-name "
                "to regenerate the full secrets mask."
            )
        else:
            _log_progress("Step 6c: Generating specs secrets mask...")
            # Reuse entries collected during Step 6a to avoid redundant GCS reads.
            secret_names = _extract_secret_property_names(all_registry_entries)
            sorted_names = sorted(secret_names)
            result.specs_secrets_mask_properties = len(sorted_names)
            mask_content = yaml.dump(
                {"properties": sorted_names}, default_flow_style=False
            )
            mask_path = f"{bucket_name}/{prefix}{_REGISTRIES_PREFIX}/{_SPECS_SECRETS_MASK_FILENAME}"

            _log_progress(
                "  Found %d secret properties: %s",
                len(sorted_names),
                ", ".join(sorted_names),
            )

            if dry_run:
                _log_progress(
                    "  [DRY RUN] Would write %s",
                    _SPECS_SECRETS_MASK_FILENAME,
                )
            else:
                with fs.open(mask_path, "w") as f:
                    f.write(mask_content)
                _log_progress(
                    "  Wrote %s",
                    _SPECS_SECRETS_MASK_FILENAME,
                )

    _log_progress(result.summary())
    return result



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Slack webhook handler for Block Kit interactive components.

Processes incoming Slack interaction payloads (button clicks, etc.),
extracts the target Devin session URL from the button value, and
injects a notification message into the session.

When an approval or rejection button is clicked, a thread reply is posted
as a durable, verifiable approval record. The reply URL is included in
the notification injected into the Devin session so that MCP tools can
use it as an `approval_comment_url`.
"""

from __future__ import annotations

import hashlib
import hmac
import json
import logging
import os
import time
from datetime import datetime, timezone
from typing import Any
from zoneinfo import ZoneInfo

import requests

from agent_message_bus.devin_client import INJECT_OK, inject_message

logger = logging.getLogger(__name__)

# Slack allows up to 5 minutes of clock skew for signature validation
_SLACK_TIMESTAMP_MAX_AGE_SECONDS = 300


def verify_slack_signature(
    payload_body: bytes,
    timestamp: str,
    signature: str,
    signing_secret: str,
) -> bool:
    """Verify a Slack request signature (X-Slack-Signature).

    Slack uses HMAC-SHA256 with a versioned signing scheme:
    `v0=HMAC_SHA256(signing_secret, "v0:{timestamp}:{body}")`

    Args:
        payload_body: Raw request body bytes.
        timestamp: Value of X-Slack-Request-Timestamp header.
        signature: Value of X-Slack-Signature header.
        signing_secret: The Slack app's signing secret.

    Returns:
        True if the signature is valid and the timestamp is recent.
    """
    if not timestamp or not signature:
        return False

    # Reject requests older than 5 minutes (replay protection)
    try:
        ts = int(timestamp)
    except ValueError:
        return False

    if abs(time.time() - ts) > _SLACK_TIMESTAMP_MAX_AGE_SECONDS:
        logger.warning("Slack request timestamp too old: %s", timestamp)
        return False

    sig_basestring = f"v0:{timestamp}:{payload_body.decode('utf-8')}"
    expected = (
        "v0="
        + hmac.new(
            signing_secret.encode("utf-8"),
            sig_basestring.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()
    )

    return hmac.compare_digest(expected, signature)


def _extract_session_url_from_action(action: dict[str, Any]) -> str | None:
    """Extract the Devin session URL from a Slack action payload.

    The session URL is expected to be embedded in the button's `value`
    field as a JSON string or a direct URL.

    Args:
        action: A single action element from the Slack interaction payload.

    Returns:
        The session URL if found, None otherwise.
    """
    value = action.get("value", "")
    if not value:
        return None

    # Try parsing as JSON first (our standard format)
    try:
        data = json.loads(value)
        if isinstance(data, dict):
            return data.get("session_url")
    except (json.JSONDecodeError, TypeError):
        pass

    # Fall back to treating the value as a direct URL
    if "app.devin.ai/sessions/" in value:
        return value

    return None


def _format_slack_notification(
    action: dict[str, Any],
    user: dict[str, Any],
    message_text: str | None = None,
    approval_reply_url: str | None = None,
) -> str:
    """Format a notification message for a Devin session from a Slack action.

    Args:
        action: The Slack action element (button click, etc.).
        user: The Slack user who performed the action.
        message_text: Optional original message text for context.
        approval_reply_url: Optional Slack message URL for the approval
            thread reply. When present, MCP tools can use this as an
            `approval_slack_url` to verify the approver's identity.

    Returns:
        Formatted notification string.
    """
    user_name = user.get("name", user.get("username", "someone"))
    action_id = action.get("action_id", "unknown_action")
    button_text = action.get("text", {}).get("text", action_id)

    parts = [
        f"Slack action received: @{user_name} clicked '{button_text}'",
    ]

    if message_text:
        # Truncate for context
        if len(message_text) > 300:
            message_text = message_text[:300] + "..."
        parts.append(f"Original message context: {message_text}")

    if approval_reply_url:
        parts.append(f"Approval record URL (for MCP tools): {approval_reply_url}")

    return "\n\n".join(parts)


# Action IDs that represent approval/rejection actions whose buttons
# should all be morphed together when any one of them is clicked.
_APPROVAL_ACTION_IDS = frozenset({"approve_request", "reject_request"})


def _update_message_after_action(
    response_url: str,
    original_message: dict[str, Any],
    clicked_action_id: str,
    user_name: str,
) -> None:
    """Update the original Slack message after an approval or rejection.

    When an approval-family button (approve or reject) is clicked, *all*
    sibling approval-family buttons are replaced with non-interactive
    status indicators so that neither can be clicked again.  Non-approval
    buttons (e.g. "View PR", "View Details") are left untouched.

    Args:
        response_url: The response_url from the Slack interaction payload.
        original_message: The original message dict from the payload.
        clicked_action_id: The action_id of the button that was clicked.
        user_name: Display name of the user who clicked.
    """
    if not response_url:
        logger.warning("No response_url in payload, cannot update message")
        return

    is_rejection = clicked_action_id == "reject_request"
    action_at = datetime.now(tz=timezone.utc).strftime("%b %-d, %Y at %H:%M UTC")

    # Compute status fields before iterating elements so they are
    # available for both the clicked button and its sibling.
    if is_rejection:
        status_emoji = ":x:"
        status_verb = "Rejected"
        dialog_title = "Action Already Rejected"
    else:
        status_emoji = ":white_check_mark:"
        status_verb = "Approved"
        dialog_title = "Action Already Approved"

    # Generic confirm body used for the sibling button (the clicked
    # button builds its own confirm_body that includes the original
    # action summary from the confirmation dialog).
    sibling_confirm_body = f"{status_verb} by @{user_name} on {action_at}"

    blocks = original_message.get("blocks", [])
    updated_blocks: list[dict[str, Any]] = []

    for block in blocks:
        if block.get("type") != "actions":
            updated_blocks.append(block)
            continue

        # Rebuild the actions block: morph all approval-family buttons
        new_elements: list[dict[str, Any]] = []
        for element in block.get("elements", []):
            element_action_id = element.get("action_id", "")

            if element_action_id not in _APPROVAL_ACTION_IDS:
                # Keep non-approval buttons (View PR, View Details, etc.)
                new_elements.append(element)
                continue

            is_clicked = element_action_id == clicked_action_id

            if is_clicked:
                # The button the user actually clicked — show full status
                original_action_summary = element.get("confirm", {}).get("text", {}).get("text", "")
                confirm_lines: list[str] = []
                if original_action_summary:
                    confirm_lines.append(original_action_summary)

                confirm_lines.append(f"{status_verb} by @{user_name} on {action_at}")
                confirm_body = "\n\n".join(confirm_lines)[:300]

                replacement: dict[str, Any] = {
                    "type": "button",
                    "text": {
                        "type": "plain_text",
                        "text": f"{status_emoji} {status_verb} by @{user_name}",
                        "emoji": True,
                    },
                    "action_id": f"{element_action_id}_done",
                    "confirm": {
                        "title": {
                            "type": "plain_text",
                            "text": dialog_title,
                        },
                        "text": {
                            "type": "mrkdwn",
                            "text": confirm_body,
                        },
                        "confirm": {
                            "type": "plain_text",
                            "text": "OK",
                        },
                        "deny": {
                            "type": "plain_text",
                            "text": "Close",
                        },
                    },
                }
                new_elements.append(replacement)
            else:
                # Sibling approval button that was NOT clicked — grey it out
                # but show the same no-op dialog so users know the outcome.
                new_elements.append(
                    {
                        "type": "button",
                        "text": {
                            "type": "plain_text",
                            "text": element.get("text", {}).get("text", element_action_id),
                            "emoji": True,
                        },
                        "action_id": f"{element_action_id}_done",
                        "confirm": {
                            "title": {
                                "type": "plain_text",
                                "text": dialog_title,
                            },
                            "text": {
                                "type": "mrkdwn",
                                "text": sibling_confirm_body,
                            },
                            "confirm": {
                                "type": "plain_text",
                                "text": "OK",
                            },
                            "deny": {
                                "type": "plain_text",
                                "text": "Close",
                            },
                        },
                    }
                )

        updated_blocks.append({"type": "actions", "elements": new_elements})

    try:
        resp = requests.post(
            response_url,
            json={
                "replace_original": True,
                "blocks": updated_blocks,
            },
            headers={"Content-Type": "application/json"},
            timeout=10,
        )
        if resp.ok:
            logger.info("Updated Slack message after %s action", clicked_action_id)
        else:
            logger.warning(
                "Failed to update Slack message via response_url: %s %s",
                resp.status_code,
                resp.text[:200],
            )
    except requests.RequestException as exc:
        logger.warning("Error posting to response_url: %s", exc)


def _get_slack_bot_token() -> str | None:
    """Retrieve the Slack bot token from environment.

    Returns:
        The bot token string, or None if not configured.
    """
    return os.environ.get("SLACK_BOT_TOKEN") or os.environ.get("SLACK_BOT_TOKEN_HITL")


def _build_slack_message_url(
    channel_id: str,
    message_ts: str,
    *,
    thread_ts: str | None = None,
) -> str:
    """Build a Slack message permalink from channel and timestamp.

    Constructs a URL in the format:
    `https://airbytehq-team.slack.com/archives/{channel}/p{ts_without_dot}`

    If *thread_ts* is provided the `?thread_ts=...&cid=...` query params
    are appended so the link opens inside the thread.

    Args:
        channel_id: Slack channel ID (e.g. `C08BHPUMEPJ`).
        message_ts: Message timestamp (e.g. `1773062711.122019`).
        thread_ts: Parent thread timestamp, if the message is a reply.

    Returns:
        Slack message URL string.
    """
    ts_no_dot = message_ts.replace(".", "")
    base = f"https://airbytehq-team.slack.com/archives/{channel_id}/p{ts_no_dot}"
    if thread_ts:
        base += f"?thread_ts={thread_ts}&cid={channel_id}"
    return base


def _post_approval_thread_reply(
    channel_id: str,
    thread_ts: str,
    user_id: str,
    user_name: str,
    is_approved: bool,
    bot_token: str,
) -> str | None:
    """Post a thread reply as a durable approval/rejection record.

    The reply contains both a human-readable summary and a structured
    metadata block that MCP tools can reliably parse to extract the
    approver's Slack user ID.

    Args:
        channel_id: Slack channel where the original message lives.
        thread_ts: Timestamp of the parent message (used as `thread_ts`).
        user_id: Slack user ID of the person who clicked the button.
        user_name: Display name of the person.
        is_approved: True for approval, False for rejection.
        bot_token: Slack Bot User OAuth Token.

    Returns:
        The Slack message URL of the posted reply, or None on failure.
    """
    action_at_utc = datetime.now(tz=timezone.utc)
    action_at = action_at_utc.strftime("%Y-%m-%dT%H:%M:%SZ")

    # Display timestamp in US/Pacific (auto-detects PST vs PDT)
    pacific_tz = ZoneInfo("America/Los_Angeles")
    action_at_pacific = action_at_utc.astimezone(pacific_tz)
    # Determine PST vs PDT label
    tz_abbr = action_at_pacific.strftime("%Z")  # e.g. "PST" or "PDT"
    if not tz_abbr or tz_abbr == action_at_pacific.strftime("%z"):
        # strftime %Z may return empty or offset on some platforms; fall back
        tz_abbr = "PT"
    action_at_display = action_at_pacific.strftime(f"%b %-d, %Y at %H:%M {tz_abbr}")

    if is_approved:
        emoji = ":white_check_mark:"
        verb = "Approved"
    else:
        emoji = ":x:"
        verb = "Rejected"

    # Human-readable line
    text = f"{emoji} *{verb}* by <@{user_id}> \u2014 _{action_at_display}_"

    # Structured metadata block for machine parsing
    metadata_json = json.dumps(
        {
            "type": "approval_record",
            "action": "approved" if is_approved else "rejected",
            "user_id": user_id,
            "user_name": user_name,
            "timestamp": action_at,
        }
    )

    blocks: list[dict[str, Any]] = [
        {
            "type": "section",
            "text": {"type": "mrkdwn", "text": text},
        },
        {
            "type": "context",
            "elements": [
                {
                    "type": "mrkdwn",
                    "text": f"```{metadata_json}```",
                }
            ],
        },
    ]

    try:
        resp = requests.post(
            "https://slack.com/api/chat.postMessage",
            headers={
                "Authorization": f"Bearer {bot_token}",
                "Content-Type": "application/json; charset=utf-8",
            },
            json={
                "channel": channel_id,
                "thread_ts": thread_ts,
                "blocks": blocks,
                "text": f"{verb} by @{user_name}",
                "unfurl_links": False,
                "unfurl_media": False,
            },
            timeout=10,
        )
        data = resp.json()
        if not data.get("ok"):
            logger.warning(
                "Slack API error posting approval reply: %s",
                data.get("error", "unknown"),
            )
            return None

        reply_ts = data.get("ts", "")
        reply_channel = data.get("channel", channel_id)
        reply_url = _build_slack_message_url(
            channel_id=reply_channel,
            message_ts=reply_ts,
            thread_ts=thread_ts,
        )
        logger.info(
            "Posted approval thread reply in %s (ts=%s): %s",
            reply_channel,
            reply_ts,
            reply_url,
        )
        return reply_url

    except requests.RequestException as exc:
        logger.warning("Error posting approval thread reply: %s", exc)
        return None


def handle_slack_interaction(payload: dict[str, Any]) -> dict[str, Any]:
    """Process a Slack interactive component payload.

    Extracts the Devin session URL from button values and injects
    a notification message into the target session. After successful
    delivery, updates the original Slack message to mark the button
    as approved.

    For approval/rejection actions, a thread reply is posted as a
    durable approval record before injecting the notification. The
    reply URL is included in the notification so MCP tools can use
    it as an `approval_comment_url`.

    Args:
        payload: The parsed Slack interaction payload.

    Returns:
        Summary dict with processing results.
    """
    payload_type = payload.get("type", "")

    if payload_type != "block_actions":
        return {"status": "skipped", "reason": f"unsupported_type: {payload_type}"}

    actions = payload.get("actions", [])
    user = payload.get("user", {})
    message = payload.get("message", {})
    message_text = message.get("text", "")
    response_url = payload.get("response_url", "")
    user_name = user.get("name", user.get("username", "someone"))
    user_id = user.get("id", "")

    # Extract channel and thread info for posting approval replies
    channel = payload.get("channel", {})
    channel_id = channel.get("id", "") if isinstance(channel, dict) else str(channel)
    message_ts = message.get("ts", "")

    bot_token = _get_slack_bot_token()

    notified = 0
    errors = 0

    for action in actions:
        action_id = action.get("action_id", "")

        # Skip post-action "done" buttons — they are non-interactive status
        # indicators created by _update_message_after_action and carry no
        # session URL.  Clicking them is harmless but should not log warnings.
        if action_id.endswith("_done"):
            logger.debug("Ignoring post-action button click: %s", action_id)
            continue

        session_url = _extract_session_url_from_action(action)
        if not session_url:
            logger.warning(
                "No session URL found in Slack action value: %s",
                action.get("value", ""),
            )
            continue

        # Post approval thread reply for approval-family actions
        approval_reply_url: str | None = None
        if action_id in _APPROVAL_ACTION_IDS and bot_token and channel_id and message_ts:
            is_approved = action_id != "reject_request"
            approval_reply_url = _post_approval_thread_reply(
                channel_id=channel_id,
                thread_ts=message_ts,
                user_id=user_id,
                user_name=user_name,
                is_approved=is_approved,
                bot_token=bot_token,
            )
        elif action_id in _APPROVAL_ACTION_IDS and not bot_token:
            logger.warning(
                "Cannot post approval thread reply: "
                "SLACK_BOT_TOKEN / SLACK_BOT_TOKEN_HITL not configured"
            )

        notification = _format_slack_notification(
            action, user, message_text, approval_reply_url=approval_reply_url
        )
        result = inject_message(session_url, notification)

        if result == INJECT_OK:
            notified += 1
            # Update the original message to reflect the action taken
            if action_id and response_url:
                _update_message_after_action(
                    response_url=response_url,
                    original_message=message,
                    clicked_action_id=action_id,
                    user_name=user_name,
                )
        else:
            errors += 1
            logger.warning(
                "Failed to inject Slack action into session %s (result=%s)",
                session_url,
                result,
            )

    return {
        "status": "processed",
        "actions_received": len(actions),
        "notified": notified,
        "errors": errors,
    }



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import importlib
from logging import Logger
from types import ModuleType
from typing import List, Optional

from connector_ops.utils import Connector  # type: ignore
from dagger import Container

BUILD_CUSTOMIZATION_MODULE_NAME = "build_customization"
BUILD_CUSTOMIZATION_SPEC_NAME = f"{BUILD_CUSTOMIZATION_MODULE_NAME}.py"
DEFAULT_MAIN_FILE_NAME = "main.py"


def get_build_customization_module(connector: Connector) -> Optional[ModuleType]:
    """Import the build_customization.py file from the connector directory if it exists.
    Returns:
        Optional[ModuleType]: The build_customization.py module if it exists, None otherwise.
    """
    build_customization_spec_path = (
        connector.code_directory / BUILD_CUSTOMIZATION_SPEC_NAME
    )

    if not build_customization_spec_path.exists() or not (
        build_customization_spec := importlib.util.spec_from_file_location(
            f"{connector.code_directory.name}_{BUILD_CUSTOMIZATION_MODULE_NAME}",
            build_customization_spec_path,
        )
    ):
        return None

    if build_customization_spec.loader is None:
        return None

    build_customization_module = importlib.util.module_from_spec(
        build_customization_spec
    )
    build_customization_spec.loader.exec_module(build_customization_module)
    return build_customization_module


def get_main_file_name(connector: Connector) -> str:
    """Get the main file name from the build_customization.py module if it exists, DEFAULT_MAIN_FILE_NAME otherwise.

    Args:
        connector (Connector): The connector to build.

    Returns:
        str: The main file name.
    """
    build_customization_module = get_build_customization_module(connector)

    return (
        build_customization_module.MAIN_FILE_NAME
        if build_customization_module
        and hasattr(build_customization_module, "MAIN_FILE_NAME")
        else DEFAULT_MAIN_FILE_NAME
    )


def get_entrypoint(connector: Connector) -> List[str]:
    main_file_name = get_main_file_name(connector)
    return ["python", f"/airbyte/integration_code/{main_file_name}"]


def apply_airbyte_entrypoint(
    connector_container: Container, connector: Connector
) -> Container:
    entrypoint = get_entrypoint(connector)

    return connector_container.with_env_variable(
        "AIRBYTE_ENTRYPOINT", " ".join(entrypoint)
    ).with_entrypoint(entrypoint)


async def pre_install_hooks(
    connector: Connector, base_container: Container, logger: Logger
) -> Container:
    """Run the pre_connector_install hook if it exists in the build_customization.py module.
    It will mutate the base_container and return it.

    Args:
        connector (Connector): The connector to build.
        base_container (Container): The base container to mutate.
        logger (Logger): The logger to use.

    Returns:
        Container: The mutated base_container.
    """
    build_customization_module = get_build_customization_module(connector)
    if build_customization_module and hasattr(
        build_customization_module, "pre_connector_install"
    ):
        base_container = await build_customization_module.pre_connector_install(
            base_container
        )
        logger.info(f"Connector {connector.technical_name} pre install hook executed.")
    return base_container


async def post_install_hooks(
    connector: Connector, connector_container: Container, logger: Logger
) -> Container:
    """Run the post_connector_install hook if it exists in the build_customization.py module.
    It will mutate the connector_container and return it.

    Args:
        connector (Connector): The connector to build.
        connector_container (Container): The connector container to mutate.
        logger (Logger): The logger to use.

    Returns:
        Container: The mutated connector_container.
    """
    build_customization_module = get_build_customization_module(connector)
    if build_customization_module and hasattr(
        build_customization_module, "post_connector_install"
    ):
        connector_container = await build_customization_module.post_connector_install(
            connector_container
        )
        logger.info(f"Connector {connector.technical_name} post install hook executed.")
    return connector_container



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from unittest import TestCase
from unittest.mock import Mock

from airbyte_cdk.models import AirbyteCatalog, AirbyteStream, SyncMode
from erd.dbml_assembler import DbmlAssembler, Source
from tests.builder import RelationshipBuilder

_A_STREAM_NAME = "a_stream_name"


class RelationshipsMergerTest(TestCase):
    def setUp(self) -> None:
        self._source = Mock(spec=Source)
        self._source.is_dynamic.return_value = False
        self._assembler = DbmlAssembler()

    def test_given_no_streams_then_database_is_empty(self) -> None:
        dbml = self._assembler.assemble(
            self._source,
            AirbyteCatalog(streams=[]),
            {"streams": [RelationshipBuilder(_A_STREAM_NAME).build()]},
        )
        assert not dbml.tables

    def test_given_stream_is_dynamic_then_ignore(self) -> None:
        self._source.is_dynamic.return_value = True
        dbml = self._assembler.assemble(
            self._source,
            AirbyteCatalog(
                streams=[
                    AirbyteStream(
                        name=_A_STREAM_NAME,
                        json_schema={"properties": {}},
                        supported_sync_modes=[SyncMode.full_refresh],
                    )
                ]
            ),
            {"streams": [RelationshipBuilder(_A_STREAM_NAME).build()]},
        )
        assert not dbml.tables

    def test_given_stream_then_populate_table(self) -> None:
        dbml = self._assembler.assemble(
            self._source,
            AirbyteCatalog(
                streams=[
                    AirbyteStream(
                        name=_A_STREAM_NAME,
                        json_schema={
                            "properties": {
                                "a_primary_key": {"type": ["null", "string"]},
                                "an_integer": {"type": ["null", "number"]},
                            }
                        },
                        supported_sync_modes=[SyncMode.full_refresh],
                        source_defined_primary_key=[["a_primary_key"]],
                    )
                ]
            ),
            {"streams": [RelationshipBuilder(_A_STREAM_NAME).build()]},
        )
        assert len(dbml.tables) == 1
        assert len(dbml.tables[0].columns) == 2
        assert dbml.tables[0].columns[0].name == "a_primary_key"
        assert dbml.tables[0].columns[0].pk
        assert dbml.tables[0].columns[1].name == "an_integer"
